/*
 * Intelbras PSO800 FXS/FXO Interface Driver for Dahdi Telephony interface
 *
 * Rewritten by Cleiber Marques da Silva <cleiber@disc-os.org>
 *
 * Based on previous works, designs, and archetectures conceived and
 * written by Mark Spencer <markster@digium.com> and Matthew Fredrickson <creslin@digium.com>.
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include "proslic.h"
#include <dahdi/wctdm_user.h>
#include "pso800.h"
/*
 *  Define for audio vs. register based ring detection
 *  
 */
 #define AUDIO_RINGCHECK
/*
 *  Define to fake a reversal polarity when the input signal
 *  is bigger than dtmfthr value (used for brazilian caller ID)
 *  
 */
#define	FAKE_POLARITY

#ifdef FAKE_POLARITY
#undef AUDIO_RINGCHECK
#endif

/*
  Experimental max loop current limit for the proslic
  Loop current limit is from 20 mA to 41 mA in steps of 3
  (according to datasheet)
  So set the value below to:
  0x00 : 20mA (default)
  0x01 : 23mA
  0x02 : 26mA
  0x03 : 29mA
  0x04 : 32mA
  0x05 : 35mA
  0x06 : 37mA
  0x07 : 41mA
*/
static int loopcurrent = 20;

static int reversepolarity = 0;

static alpha  indirect_regs[] =
{
{0,255,"DTMF_ROW_0_PEAK",0x55C2},
{1,255,"DTMF_ROW_1_PEAK",0x51E6},
{2,255,"DTMF_ROW2_PEAK",0x4B85},
{3,255,"DTMF_ROW3_PEAK",0x4937},
{4,255,"DTMF_COL1_PEAK",0x3333},
{5,255,"DTMF_FWD_TWIST",0x0202},
{6,255,"DTMF_RVS_TWIST",0x0202},
{7,255,"DTMF_ROW_RATIO_TRES",0x0198},
{8,255,"DTMF_COL_RATIO_TRES",0x0198},
{9,255,"DTMF_ROW_2ND_ARM",0x0611},
{10,255,"DTMF_COL_2ND_ARM",0x0202},
{11,255,"DTMF_PWR_MIN_TRES",0x00E5},
{12,255,"DTMF_OT_LIM_TRES",0x0A1C},
{13,0,"OSC1_COEF",0x7B30},
{14,1,"OSC1X",0x0063},
{15,2,"OSC1Y",0x0000},
{16,3,"OSC2_COEF",0x7870},
{17,4,"OSC2X",0x007D},
{18,5,"OSC2Y",0x0000},
{19,6,"RING_V_OFF",0x0000},
{20,7,"RING_OSC",0x7EF0},
{21,8,"RING_X",0x0160},
{22,9,"RING_Y",0x0000},
{23,255,"PULSE_ENVEL",0x2000},
{24,255,"PULSE_X",0x2000},
{25,255,"PULSE_Y",0x0000},
//{26,13,"RECV_DIGITAL_GAIN",0x4000},	// playback volume set lower
{26,13,"RECV_DIGITAL_GAIN",0x2000},	// playback volume set lower
{27,14,"XMIT_DIGITAL_GAIN",0x4000},
//{27,14,"XMIT_DIGITAL_GAIN",0x2000},
{28,15,"LOOP_CLOSE_TRES",0x1000},
{29,16,"RING_TRIP_TRES",0x3600},
{30,17,"COMMON_MIN_TRES",0x1000},
{31,18,"COMMON_MAX_TRES",0x0200},
{32,19,"PWR_ALARM_Q1Q2",0x07C0},
{33,20,"PWR_ALARM_Q3Q4",0x2600},
{34,21,"PWR_ALARM_Q5Q6",0x1B80},
{35,22,"LOOP_CLOSURE_FILTER",0x8000},
{36,23,"RING_TRIP_FILTER",0x0320},
{37,24,"TERM_LP_POLE_Q1Q2",0x008C},
{38,25,"TERM_LP_POLE_Q3Q4",0x0100},
{39,26,"TERM_LP_POLE_Q5Q6",0x0010},
{40,27,"CM_BIAS_RINGING",0x0C00},
{41,64,"DCDC_MIN_V",0x0C00},
{42,255,"DCDC_XTRA",0x1000},
{43,66,"LOOP_CLOSE_TRES_LOW",0x1000},
};

static struct fxo_mode {
	char *name;
	/* FXO */
	int ohs;
	int ohs2;
	int gain_rx;
	int rz;
	int rt;
	int rt2;
	int ilim;
	int dcv;
	int mini;
	int acim;
	int ring_osc;
	int ring_x;
} fxo_modes[] =
{
	{ "FCC", 0, 0, 0, 0, 1, 0, 0, 0x3, 0, 0, }, 	/* US, Canada */
	{ "TBR21", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, 0x7e6c, 0x023a, },
										/* Austria, Belgium, Denmark, Finland, France, Germany, 
										   Greece, Iceland, Ireland, Italy, Luxembourg, Netherlands,
										   Norway, Portugal, Spain, Sweden, Switzerland, and UK */
	{ "ARGENTINA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "AUSTRALIA", 1, 0, 0, 0, 0, 0, 0, 0, 0x3, 0x3, },
	{ "AUSTRIA", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x3, },
	{ "BAHRAIN", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "BELGIUM", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	/* { "BRAZIL", 0, 0, 0, 0, 0, 0, 0x3, 0, },*/
	{ "BRAZIL", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 1, },
	{ "BULGARIA", 0, 0, 0, 0, 0, 0, 1, 0x3, 0x0, 0x3, },
	{ "CANADA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "CHILE", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "CHINA", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0xf, },
	{ "COLUMBIA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "CROATIA", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "CYRPUS", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "CZECH", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "DENMARK", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "ECUADOR", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "EGYPT", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "ELSALVADOR", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "FINLAND", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "FRANCE", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "GERMANY", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x3, },
	{ "GREECE", 0, 1, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "GUAM", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "HONGKONG", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "HUNGARY", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "ICELAND", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "INDIA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0x4, },
	{ "INDONESIA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "IRELAND", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "ISRAEL", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "ITALY", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "JAPAN", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "JORDAN", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "KAZAKHSTAN", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "KUWAIT", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "LATVIA", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "LEBANON", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "LUXEMBOURG", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "MACAO", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "MALAYSIA", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },	/* Current loop >= 20ma */
	{ "MALTA", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "MEXICO", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "MOROCCO", 0, 0, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "NETHERLANDS", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "NEWZEALAND", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0x4, },
	{ "NIGERIA", 0, 0, 0, 0, 0, 0, 0x1, 0x3, 0, 0x2, },
	{ "NORWAY", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "OMAN", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "PAKISTAN", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "PERU", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "PHILIPPINES", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "POLAND", 0, 0, 0, 1, 1, 0, 0, 0x3, 0, 0, },
	{ "PORTUGAL", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "ROMANIA", 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, },
	{ "RUSSIA", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "SAUDIARABIA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "SINGAPORE", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "SLOVAKIA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0x3, },
	{ "SLOVENIA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0x2, },
	{ "SOUTHAFRICA", 1, 0, 0, 1, 0, 0, 0, 0x3, 0, 0x3, },
	{ "SOUTHKOREA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "SPAIN", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "SWEDEN", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "SWITZERLAND", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "SYRIA", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "TAIWAN", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "THAILAND", 0, 0, 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "UAE", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "UK", 0, 1, 0, 0, 0, 0, 1, 0x3, 0, 0x5, },
	{ "USA", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "YEMEN", 0, 0, 0, 0, 0, 0, 0, 0x3, 0, 0, },
};

#include <dahdi/kernel.h>

#include <linux/moduleparam.h>

#define NUM_FXO_REGS 60

#define WC_MAX_IFACES 128

#define WC_CARD_TYPE_REG	0x01
#define WC_LEDS_REG			0x02
#define WC_SPI_REG 			0x03
#define WC_RST_REG 			0x05
#define WC_CNT_IRQ_REG		0x06
#define WC_PICO_REG			0x07
#define WC_CLEAR_IRQ_REG	0x0C
#define WC_DATA_BASE		0x0100


#define BIT_SDI		(1 << 0)
#define BIT_SDO		(1 << 1)
#define BIT_SCLK	(1 << 2)
#define BIT_CS		(1 << 3)

#define WC_LED_OFF		0
#define WC_LED_ON		1
#define WC_LED_BLINK	2


/* the constants below control the 'debounce' periods enforced by the
   check_hook routines; these routines are called once every 4 interrupts
   (the interrupt cycles around the four modules), so the periods are
   specified in _4 millisecond_ increments
*/
#define RING_DEBOUNCE		4		/* Ringer Debounce (64 ms) */
#define DEFAULT_BATT_DEBOUNCE	4		/* Battery debounce (64 ms) */
#define POLARITY_DEBOUNCE 	4		/* Polarity debounce (64 ms) */
#define DEFAULT_BATT_THRESH	3		/* Anything under this is "no battery" */

#define OHT_TIMER		6000	/* How long after RING to retain OHT */

#define FLAG_3215	(1 << 0)

#define NUM_CARDS 8
#define NUM_MOD_CHS 4

#define MAX_ALARMS 10

#define MOD_TYPE_NONE	0
#define MOD_TYPE_FXS	1
#define MOD_TYPE_FXO	2

#define MINPEGTIME	10 * 8		/* 30 ms peak to peak gets us no more than 100 Hz */
#define PEGTIME		50 * 8		/* 50ms peak to peak gets us rings of 10 Hz or more */
#define PEGCOUNT	5		/* 5 cycles of pegging means RING */

#define NUM_CAL_REGS 12

//#define pd3_debugk(fmt,args...) printk("pd3 -> %s: "fmt, __PRETTY_FUNCTION__, ##args)
#define pd3_debugk(fmt,args...) 

#define PCI_DEVICE_ID_PLX_9030	0x9030

struct fxo_mode channels[NUM_CARDS];

struct calregs {
	unsigned char vals[NUM_CAL_REGS];
};

enum proslic_power_warn {
	PROSLIC_POWER_UNKNOWN = 0,
	PROSLIC_POWER_ON,
	PROSLIC_POWER_WARNED,
};

struct pso800 {
	struct pci_dev *dev;
	char *variety;
	struct dahdi_span span;
	unsigned int ios;
	int usecount;
	unsigned int intcount;
	int flag_1st_irq;
	int fpgaver;
	unsigned char led[NUM_CARDS];
	char proc_filename[64];
	unsigned int write_key;
	int dead;
	int pos;
	int flags[NUM_CARDS];
	int freeregion;
	int alt;
	int curcard;
	int cardflag;		/* Bit-map of present cards */
	enum proslic_power_warn proslic_power;
	spinlock_t lock;

	union {
		struct {
#ifdef AUDIO_RINGCHECK
			unsigned int pegtimer;
			int pegcount;
			int peg;
			int ring;
#else			
			int wasringing;
#endif			
			int ringdebounce;
			int offhook;
			int battdebounce;
			int nobatttimer;
			int battery;
			int lastpol;
			int polarity;
			int polaritydebounce;
#ifdef FAKE_POLARITY
			int readcid;
			unsigned int cidtimer;
			int cidflag;
			unsigned int ciddebounce;
#endif					
		} fxo;
		struct {
			int oldrxhook;
			int debouncehook;
			int lastrxhook;
			int debounce;
			int ohttimer;
			int idletxhookstate;		/* IDLE changing hook state */
			int lasttxhook;
			int palarms;
			struct calregs calregs;
		} fxs;
	} mod[NUM_CARDS];

	/* Receive hook state and debouncing */
	int modtype[NUM_CARDS];
	unsigned char reg0shadow[NUM_CARDS];
	unsigned char reg1shadow[NUM_CARDS];

	unsigned long memaddr;
	unsigned long memlen;
	volatile unsigned int *ioaddr;
	volatile unsigned int *pci_vm_addr;

	dma_addr_t 	readdma;
	dma_addr_t	writedma;
	volatile unsigned int *writechunk;				/* Double-word aligned write memory */
	volatile unsigned int *readchunk;				/* Double-word aligned read memory */
	struct dahdi_chan _chans[NUM_CARDS];
	struct dahdi_chan *chans[NUM_CARDS];
};


struct pso800_desc {
	char *name;
	int flags;
};

static struct pso800_desc pso800r1 = { "Intelbras PSO800", 0 };

static int acim2tiss[16] = { 0x0, 0x1, 0x4, 0x5, 0x7, 0x0, 0x0, 0x6, 0x0, 0x0, 0x0, 0x2, 0x0, 0x3 };

static struct pso800 *ifaces[WC_MAX_IFACES];

static void pso800_release(struct pso800 *wc);

static int battdebounce = DEFAULT_BATT_DEBOUNCE;
static int battthresh = DEFAULT_BATT_THRESH;
static int debug = 0;
static int robust = 0;
static int timingonly = 0;
static int lowpower = 0;
static int boostringer = 0;
static int fastringer = 0;
static int _opermode = 0;
static char *opermode = "BRAZIL";//"FCC";
/* XXX UGLY XXX */
static char *chan1 = NULL, *chan2 = NULL, *chan3 = NULL, *chan4 = NULL, *chan5 = NULL, *chan6 = NULL, *chan7 = NULL, *chan8 = NULL;
static int fxshonormode = 0;
static int alawoverride = 1;
#ifdef FAKE_POLARITY
/*	define this between 2000 and 20000. The value of 2000 is the minimum level at the input
	that must be recognized as a dtmf signal and represents something around -25dBm. Also,
	the dtfm signal should be less than 20000, that represents something like -3dBm */
static int dtmfthr = 4000;
static int minvoltage = 18;
#endif

static int pso800_init_proslic(struct pso800 *wc, int card, int fast , int manual, int sane);

unsigned int pso800_get_serial(struct pci_dev *dev, struct pso800 *wc);
unsigned int pso800_e2p_date(struct pci_dev *dev, struct pso800 *wc);

static void wait_just_a_bit(int foo)
{
	long newjiffies;
	newjiffies = jiffies + foo;
	while(jiffies < newjiffies);
}

static inline void __pso800_set_led(struct pso800 *wc, int led)
{
	unsigned int ledreg = *(wc->ioaddr+WC_LEDS_REG);
	
	ledreg |= (1<<led);
	*(wc->ioaddr+WC_LEDS_REG) = ledreg;
}

static inline void __pso800_clear_led(struct pso800 *wc, int led)
{
	unsigned int ledreg = *(wc->ioaddr+WC_LEDS_REG);
	
	ledreg &= ~(1<<led);
	*(wc->ioaddr+WC_LEDS_REG) = ledreg;
}

static inline void __pso800_blink_led(struct pso800 *wc, int led)
{
	unsigned int ledreg = *(wc->ioaddr+WC_LEDS_REG);
	
	if(ledreg & (1<<led))
		ledreg &= ~(1<<led);
	else
		ledreg |= (1<<led);
	
	*(wc->ioaddr+WC_LEDS_REG) = ledreg;
}

static inline void pso800_transmitprep(struct pso800 *wc)
{
	volatile unsigned int *writechunk;
	unsigned int writesample;
	int x, z;
		
	writechunk = wc->ioaddr + WC_DATA_BASE;

	/* Calculate Transmission */
	dahdi_transmit(&wc->span);

	for (x=0;x<DAHDI_CHUNKSIZE;x++) {	
		for (z=0;z<4;z++) {		
#ifdef __BIG_ENDIAN			
			writesample = (wc->chans[0+z]->writechunk[x]<<24) | (wc->chans[4+z]->writechunk[x] << 16);
#else			
			writesample = (wc->chans[0+z]->writechunk[x]) | (wc->chans[4+z]->writechunk[x] << 8);
#endif
			writechunk[z] = writesample;
		}
		writechunk += 4;		
	}

}

#ifdef AUDIO_RINGCHECK
static inline void ring_check(struct pso800 *wc, int card)
{
	int x;
	short sample;
	if (wc->modtype[card] != MOD_TYPE_FXO)
		return;
	wc->mod[card].fxo.pegtimer += DAHDI_CHUNKSIZE;
	for (x=0;x<DAHDI_CHUNKSIZE;x++) {
		/* Look for pegging to indicate ringing */
		sample = DAHDI_XLAW(wc->chans[card]->readchunk[x], (wc->chans[card]));
		if ((sample > 10000) && (wc->mod[card].fxo.peg != 1)) {
			if (debug > 1) printk("High peg!\n");
			if ((wc->mod[card].fxo.pegtimer < PEGTIME) && (wc->mod[card].fxo.pegtimer > MINPEGTIME))
				wc->mod[card].fxo.pegcount++;
			wc->mod[card].fxo.pegtimer = 0;
			wc->mod[card].fxo.peg = 1;
		} else if ((sample < -10000) && (wc->mod[card].fxo.peg != -1)) {
			if (debug > 1) printk("Low peg!\n");
			if ((wc->mod[card].fxo.pegtimer < (PEGTIME >> 2)) && (wc->mod[card].fxo.pegtimer > (MINPEGTIME >> 2)))
				wc->mod[card].fxo.pegcount++;
			wc->mod[card].fxo.pegtimer = 0;
			wc->mod[card].fxo.peg = -1;
		}
	}
	if (wc->mod[card].fxo.pegtimer > PEGTIME) {
		/* Reset pegcount if our timer expires */
		wc->mod[card].fxo.pegcount = 0;
	}
	/* Decrement debouncer if appropriate */
	if (wc->mod[card].fxo.ringdebounce)
		wc->mod[card].fxo.ringdebounce--;
	if (!wc->mod[card].fxo.offhook && !wc->mod[card].fxo.ringdebounce) {
		if (!wc->mod[card].fxo.ring && (wc->mod[card].fxo.pegcount > PEGCOUNT)) {
			/* It's ringing */
			if (debug)
				printk("RING on %d/%d!\n", wc->span.spanno, card + 1);
			if (!wc->mod[card].fxo.offhook)
				dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_RING);
			wc->mod[card].fxo.ring = 1;
		}
		if (wc->mod[card].fxo.ring && !wc->mod[card].fxo.pegcount) {
			/* No more ring */
			if (debug)
				printk("NO RING on %d/%d!\n", wc->span.spanno, card + 1);
			dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
			wc->mod[card].fxo.ring = 0;
		}
	}
}
#endif

#ifdef FAKE_POLARITY
static inline void pso800_dtmfcheck_fakepolarity(struct pso800 *wc, int card) 
{
	int x, sample;
	
/*	static int max_sample = 0;	

	for (x=0;x<DAHDI_CHUNKSIZE;x++) {
	   	sample = DAHDI_XLAW(wc->chans[card].readchunk[x], (&(wc->chans[card])));
		if(wc->intcount > 4000)
			if(abs(sample) > abs(max_sample)) max_sample = sample;
		if(wc->intcount > wc->mod[card].fxo.cidtimer) {
			if (debug && card==1) printk("DTMF sample(%d) = %d on %d (intcount = %d)\n", x, sample, card+1, wc->intcount);
			if((x==7) && (card==1) && (wc->intcount > (wc->mod[card].fxo.cidtimer+1))) {
				wc->mod[card].fxo.cidtimer = wc->intcount+2000;
				printk("max_sample = %d\n\n", max_sample);
			}
		}
	}
	return;
*/
	for (x=0;x<DAHDI_CHUNKSIZE;x++) {
	
		sample = DAHDI_XLAW(wc->chans[card]->readchunk[x], ((wc->chans[card])));
		
		/*if(!(wc->intcount%2000)) {
			if (debug) printk("chan = %i, sample(%d) = %d\n",card+1, x, sample);
		}*/

	    /* don't look for noise if we're already processing it, or there is a ringing tone */
		if(!wc->mod[card].fxo.readcid && !wc->mod[card].fxo.wasringing  && (wc->intcount > (wc->mod[card].fxo.cidtimer + 500)) ) {
			if(!wc->mod[card].fxo.cidflag) {
				if (abs(sample) > dtmfthr) {
					wc->mod[card].fxo.cidflag = 1;
					wc->mod[card].fxo.ciddebounce = wc->intcount;
					if (debug) printk("cidflag = 1 on %i (sample = %d, intcount = %d)\n",card+1, sample, wc->intcount);
				}
			} else if(wc->intcount > (wc->mod[card].fxo.ciddebounce + 4)) {
				if(wc->intcount < (wc->mod[card].fxo.ciddebounce + 9)) {
					if (abs(sample) > dtmfthr) {
						wc->mod[card].fxo.readcid = 1;
						wc->mod[card].fxo.cidtimer = wc->intcount;
						if (debug) printk("DTMF CLIP on %i (sample = %d, intcount = %d)\n",card+1, sample, wc->intcount);
						dahdi_qevent_lock(wc->chans[card], DAHDI_EVENT_POLARITY);
					}
				} else {
					wc->mod[card].fxo.cidflag = 0;
					if (debug) printk("cidflag = 0, it was noise on %i (sample = %d, intcount = %d)\n",card+1, sample, wc->intcount);
				}
			}
		} else if(wc->mod[card].fxo.readcid && (wc->intcount > (wc->mod[card].fxo.cidtimer + 4000))) {
			/* reset flags if it's been a while */
			wc->mod[card].fxo.cidtimer = wc->intcount;
			wc->mod[card].fxo.readcid = 0;
			wc->mod[card].fxo.cidflag = 0;
			if (debug) printk("readcid = 0 on %d (sample = %d, intcount = %d)\n",card+1, sample, wc->intcount);
		}
	}
}
#endif

static inline void pso800_receiveprep(struct pso800 *wc)
{
	volatile unsigned int *readchunk;
	unsigned int readsample;
	int x, z;
#ifdef FAKE_POLARITY
	signed char line_voltage;
#endif

	readchunk = wc->ioaddr + WC_DATA_BASE;

	for (x=0;x<DAHDI_CHUNKSIZE;x++) {	
		for (z=0;z<4;z++) {
			readsample = readchunk[z];
#ifdef __BIG_ENDIAN
			wc->chans[4+z].readchunk[x] = (readsample >> 16) & 0xff;
			wc->chans[0+z].readchunk[x] = (readsample >> 24) & 0xff;
#else			
			wc->chans[0+z]->readchunk[x] = (readsample) & 0xff;
			wc->chans[4+z]->readchunk[x] = (readsample >> 8) & 0xff;
#endif
		}
		readchunk += 4;		
	}
			

#ifdef AUDIO_RINGCHECK
	for (x=0; x < NUM_CARDS; x++)
		ring_check(wc, x);
#endif

#ifdef FAKE_POLARITY
	for (x = 0; x < NUM_CARDS; x++) {
		if (!(wc->cardflag & (1 << x)) || !(wc->modtype[x] == MOD_TYPE_FXO) || wc->mod[x].fxo.offhook ) 
			continue;
		line_voltage = (signed char)wc->reg1shadow[x];
		if (abs(line_voltage) > minvoltage)
			pso800_dtmfcheck_fakepolarity(wc, x);
	}
#endif

	/* XXX We're wasting 8 taps.  We should get closer :( */
	for (x = 0; x < NUM_CARDS; x++) {
		if (wc->cardflag & (1 << x))
			dahdi_ec_chunk(wc->chans[x], wc->chans[x]->readchunk, wc->chans[x]->writechunk);
	}
	dahdi_receive(&wc->span);
}


static inline void __write_8bits(struct pso800 *wc, unsigned char bits)
{
	int x;
		
	wc->ios |= BIT_SCLK;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	wc->ios &= ~BIT_CS;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	for (x=0;x<8;x++) {
		/* Send out each bit, MSB first, drop SCLK as we do so */
		if (bits & 0x80)
			wc->ios |= BIT_SDI;
		else
			wc->ios &= ~BIT_SDI;
		wc->ios &= ~BIT_SCLK;
		*(wc->ioaddr + WC_SPI_REG) = wc->ios;
		/* Now raise SCLK high again and repeat */
		wc->ios |= BIT_SCLK;
		*(wc->ioaddr + WC_SPI_REG) = wc->ios;
		bits <<= 1;
	}
	/* Finally raise CS back high again */
	wc->ios |= BIT_CS;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	
}

static inline void __reset_spi(struct pso800 *wc)
{
	/* Drop chip select and clock once and raise and clock once */
	wc->ios |= BIT_SCLK;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	wc->ios &= ~BIT_CS;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	wc->ios |= BIT_SDI;
	wc->ios &= ~BIT_SCLK;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	/* Now raise SCLK high again and repeat */
	wc->ios |= BIT_SCLK;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	/* Finally raise CS back high again */
	wc->ios |= BIT_CS;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	/* Clock again */
	wc->ios &= ~BIT_SCLK;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	/* Now raise SCLK high again and repeat */
	wc->ios |= BIT_SCLK;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	
}

static inline unsigned char __read_8bits(struct pso800 *wc)
{
	unsigned char res=0, c;
	int x;
	wc->ios |= BIT_SCLK;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	/* Drop chip select */
	wc->ios &= ~BIT_CS;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	for (x=0;x<8;x++) {
		res <<= 1;
		/* Get SCLK */
		wc->ios &= ~BIT_SCLK;
		*(wc->ioaddr + WC_SPI_REG) = wc->ios;
		/* Read back the value */
		c = (unsigned char)(*(wc->ioaddr + WC_SPI_REG));
		if (c & BIT_SDO)
			res |= 1;
		/* Now raise SCLK high again */
		wc->ios |= BIT_SCLK;
		*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	}
	/* Finally raise CS back high again */
	wc->ios |= BIT_CS;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;
	wc->ios &= ~BIT_SCLK;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;

	/* And return our result */
	return res;
}


static inline void __pso800_setcard(struct pso800 *wc, int card)
{
	wc->ios &= ~0x00000f00;
	wc->ios |= card<<8;
	wc->curcard = card;
}

static void __pso800_setreg(struct pso800 *wc, int card, unsigned char reg, unsigned char value)
{
	__pso800_setcard(wc, card);
	if (wc->modtype[card] == MOD_TYPE_FXO) {
		__write_8bits(wc, 0x20);
		__write_8bits(wc, reg & 0x7f);
	} else {
		__write_8bits(wc, reg & 0x7f);
	}
	__write_8bits(wc, value);
}

static void pso800_setreg(struct pso800 *wc, int card, unsigned char reg, unsigned char value)
{
	unsigned long flags;
	spin_lock_irqsave(&wc->lock, flags);
	__pso800_setreg(wc, card, reg, value);
	spin_unlock_irqrestore(&wc->lock, flags);
}

static unsigned char __pso800_getreg(struct pso800 *wc, int card, unsigned char reg)
{
	__pso800_setcard(wc, card);
	if (wc->modtype[card] == MOD_TYPE_FXO) {
		__write_8bits(wc, 0x60);
		__write_8bits(wc, reg & 0x7f);
	} else {
		__write_8bits(wc, reg | 0x80);
	}
	return __read_8bits(wc);
}

static inline void reset_spi(struct pso800 *wc, int card)
{
	unsigned long flags;
	spin_lock_irqsave(&wc->lock, flags);
	__pso800_setcard(wc, card);
	__reset_spi(wc);
	__reset_spi(wc);
	spin_unlock_irqrestore(&wc->lock, flags);
}

static unsigned char pso800_getreg(struct pso800 *wc, int card, unsigned char reg)
{
	unsigned long flags;
	unsigned char res;
	spin_lock_irqsave(&wc->lock, flags);
	res = __pso800_getreg(wc, card, reg);
	spin_unlock_irqrestore(&wc->lock, flags);
	return res;
}

static int __wait_access(struct pso800 *wc, int card)
{
    unsigned char data = 0;
    long origjiffies;
    int count = 0;

    #define MAX 6000 /* attempts */


    origjiffies = jiffies;
    /* Wait for indirect access */
    while (count++ < MAX)
	 {
		data = __pso800_getreg(wc, card, I_STATUS);

		if (!data)
			return 0;

	 }

    if(count > (MAX-1)) printk(" ##### Loop error (%02x) #####\n", data);

	return 0;
}

static unsigned char translate_3215(unsigned char address)
{
	int x;
	for (x=0;x<sizeof(indirect_regs)/sizeof(indirect_regs[0]);x++) {
		if (indirect_regs[x].address == address) {
			address = indirect_regs[x].altaddr;
			break;
		}
	}
	return address;
}

static int pso800_proslic_setreg_indirect(struct pso800 *wc, int card, unsigned char address, unsigned short data)
{
	unsigned long flags;
	int res = -1;
	/* Translate 3215 addresses */
	if (wc->flags[card] & FLAG_3215) {
		address = translate_3215(address);
		if (address == 255)
			return 0;
	}
	spin_lock_irqsave(&wc->lock, flags);
	if(!__wait_access(wc, card)) {
		__pso800_setreg(wc, card, IDA_LO,(unsigned char)(data & 0xFF));
		__pso800_setreg(wc, card, IDA_HI,(unsigned char)((data & 0xFF00)>>8));
		__pso800_setreg(wc, card, IAA,address);
		res = 0;
	};
	spin_unlock_irqrestore(&wc->lock, flags);
	return res;
}

static int pso800_proslic_getreg_indirect(struct pso800 *wc, int card, unsigned char address)
{ 
	unsigned long flags;
	int res = -1;
	char *p=NULL;
	/* Translate 3215 addresses */
	if (wc->flags[card] & FLAG_3215) {
		address = translate_3215(address);
		if (address == 255)
			return 0;
	}
	spin_lock_irqsave(&wc->lock, flags);
	if (!__wait_access(wc, card)) {
		__pso800_setreg(wc, card, IAA, address);
		if (!__wait_access(wc, card)) {
			unsigned char data1, data2;
			data1 = __pso800_getreg(wc, card, IDA_LO);
			data2 = __pso800_getreg(wc, card, IDA_HI);
			res = data1 | (data2 << 8);
		} else
			p = "Failed to wait inside\n";
	} else
		p = "failed to wait\n";
	spin_unlock_irqrestore(&wc->lock, flags);
	if (p)
		printk(p);
	return res;
}

static int pso800_proslic_init_indirect_regs(struct pso800 *wc, int card)
{
	unsigned char i;

	for (i=0; i<sizeof(indirect_regs) / sizeof(indirect_regs[0]); i++)
	{
		if(pso800_proslic_setreg_indirect(wc, card, indirect_regs[i].address,indirect_regs[i].initial))
			return -1;
	}

	return 0;
}

static int pso800_proslic_verify_indirect_regs(struct pso800 *wc, int card)
{ 
	int passed = 1;
	unsigned short i, initial;
	int j;

	for (i=0; i<sizeof(indirect_regs) / sizeof(indirect_regs[0]); i++) 
	{
		if((j = pso800_proslic_getreg_indirect(wc, card, (unsigned char) indirect_regs[i].address)) < 0) {
			printk("Failed to read indirect register %d\n", i);
			return -1;
		}
		initial= indirect_regs[i].initial;

		if ( j != initial && (!(wc->flags[card] & FLAG_3215) || (indirect_regs[i].altaddr != 255)))
		{
			 printk("!!!!!!! %s  iREG %X = %X  should be %X\n",
				indirect_regs[i].name,indirect_regs[i].address,j,initial );
			 passed = 0;
		}	
	}

    if (passed) {
		if (debug)
			printk("Init Indirect Registers completed successfully.\n");
    } else {
		printk(" !!!!! Init Indirect Registers UNSUCCESSFULLY.\n");
		return -1;
    }
    return 0;
}

static inline void pso800_proslic_recheck_sanity(struct pso800 *wc, int card)
{
	int res;
	/* Check loopback */
	res = wc->reg1shadow[card];
	if (!res && (res != wc->mod[card].fxs.lasttxhook)) {
		res = pso800_getreg(wc, card, 8);
		if (res) {
			printk("Ouch, part reset, quickly restoring reality (%d)\n", card);
			pso800_init_proslic(wc, card, 1, 0, 1);
		} else {
			if (wc->mod[card].fxs.palarms++ < MAX_ALARMS) {
				printk("Power alarm on module %d, resetting!\n", card + 1);
				if (wc->mod[card].fxs.lasttxhook == 4)
					wc->mod[card].fxs.lasttxhook = 1;
				pso800_setreg(wc, card, 64, wc->mod[card].fxs.lasttxhook);
			} else {
				if (wc->mod[card].fxs.palarms == MAX_ALARMS)
					printk("Too many power alarms on card %d, NOT resetting!\n", card + 1);
			}
		}
	}
}

static inline void pso800_voicedaa_check_hook(struct pso800 *wc, int card)
{
#ifndef AUDIO_RINGCHECK
	unsigned char res;
#endif	
	signed char b;
	int poopy = 0;
	/* Try to track issues that plague slot one FXO's */
	b = wc->reg0shadow[card];
	if ((b & 0x2) || !(b & 0x8)) {
		/* Not good -- don't look at anything else */
		if (debug)
			printk("Poopy (%02x) on card %d!\n", b, card + 1); 
		poopy++;
	}
	b &= 0x9b;
	if (wc->mod[card].fxo.offhook) {
		if (b != 0x9)
			pso800_setreg(wc, card, 5, 0x9);
	} else {
		if (b != 0x8)
			pso800_setreg(wc, card, 5, 0x8);
	}
	if (poopy)
		return;
#ifndef AUDIO_RINGCHECK
	if (!wc->mod[card].fxo.offhook) {
		res = wc->reg0shadow[card];
		if ((res & 0x60) && wc->mod[card].fxo.battery) {
			wc->mod[card].fxo.ringdebounce += (DAHDI_CHUNKSIZE * 16);
			if (wc->mod[card].fxo.ringdebounce >= DAHDI_CHUNKSIZE * 64) {
				if (!wc->mod[card].fxo.wasringing) {
					wc->mod[card].fxo.wasringing = 1;
					dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_RING);
					if (debug)
						printk("RING on %d/%d!\n", wc->span.spanno, card + 1);
				}
				wc->mod[card].fxo.ringdebounce = DAHDI_CHUNKSIZE * 64;
			}
		} else {
			wc->mod[card].fxo.ringdebounce -= DAHDI_CHUNKSIZE * 4;
			if (wc->mod[card].fxo.ringdebounce <= 0) {
				if (wc->mod[card].fxo.wasringing) {
					wc->mod[card].fxo.wasringing = 0;
#ifdef FAKE_POLARITY
					wc->mod[card].fxo.readcid = 0;
					wc->mod[card].fxo.cidflag = 0;
					wc->mod[card].fxo.cidtimer = wc->intcount;
#endif
					dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
					if (debug)
						printk("NO RING on %d/%d!\n", wc->span.spanno, card + 1);
				}
				wc->mod[card].fxo.ringdebounce = 0;
			}
				
		}
	}
#endif
	b = wc->reg1shadow[card];
#if 0 
	{
		static int count = 0;
		if (!(count++ % 100)) {
			printk("Card %d: Voltage: %d  Debounce %d\n", card + 1, 
			       b, wc->mod[card].fxo.battdebounce);
		}
	}
#endif	
	if (abs(b) < battthresh) {
		wc->mod[card].fxo.nobatttimer++;
#if 0
		if (wc->mod[card].fxo.battery)
			printk("Battery loss: %d (%d debounce)\n", b, wc->mod[card].fxo.battdebounce);
#endif
		if (wc->mod[card].fxo.battery && !wc->mod[card].fxo.battdebounce) {
			if (debug)
				printk("NO BATTERY on %d/%d!\n", wc->span.spanno, card + 1);
			wc->mod[card].fxo.battery =  0;
#ifdef	JAPAN
			if ((!wc->ohdebounce) && wc->offhook) {
				dahdi_hooksig(&wc->chans[card], DAHDI_RXSIG_ONHOOK);
				if (debug)
					printk("Signalled On Hook\n");
#ifdef	ZERO_BATT_RING
				wc->onhook++;
#endif
			}
#else
			dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_ONHOOK);
#endif
			wc->mod[card].fxo.battdebounce = battdebounce;
		} else if (!wc->mod[card].fxo.battery)
			wc->mod[card].fxo.battdebounce = battdebounce;
	} else if (abs(b) > battthresh) {
		if (!wc->mod[card].fxo.battery && !wc->mod[card].fxo.battdebounce) {
			if (debug)
				printk("BATTERY on %d/%d (%s)!\n", wc->span.spanno, card + 1, 
					(b < 0) ? "-" : "+");			    
#ifdef	ZERO_BATT_RING
			if (wc->onhook) {
				wc->onhook = 0;
				dahdi_hooksig(&wc->chans[card], DAHDI_RXSIG_OFFHOOK);
				if (debug)
					printk("Signalled Off Hook\n");
			}
#else
			dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
#endif
			wc->mod[card].fxo.battery = 1;
			wc->mod[card].fxo.nobatttimer = 0;
			wc->mod[card].fxo.battdebounce = battdebounce;
		} else if (wc->mod[card].fxo.battery)
			wc->mod[card].fxo.battdebounce = battdebounce;

		if (wc->mod[card].fxo.lastpol >= 0) {
		    if (b < 0) {
				wc->mod[card].fxo.lastpol = -1;
				wc->mod[card].fxo.polaritydebounce = POLARITY_DEBOUNCE;
		    }
		} 
		if (wc->mod[card].fxo.lastpol <= 0) {
		    if (b > 0) {
				wc->mod[card].fxo.lastpol = 1;
				wc->mod[card].fxo.polaritydebounce = POLARITY_DEBOUNCE;
		    }
		}
	} else {
		/* It's something else... */
		wc->mod[card].fxo.battdebounce = battdebounce;
	}
	if (wc->mod[card].fxo.battdebounce)
		wc->mod[card].fxo.battdebounce--;
	if (wc->mod[card].fxo.polaritydebounce) {
	    wc->mod[card].fxo.polaritydebounce--;
		if (wc->mod[card].fxo.polaritydebounce < 1) {
		    if (wc->mod[card].fxo.lastpol != wc->mod[card].fxo.polarity) {
				if (debug)
					printk("%lu Polarity reversed on %d (%d -> %d)\n", jiffies, card+1, wc->mod[card].fxo.polarity, wc->mod[card].fxo.lastpol);
				if (wc->mod[card].fxo.polarity)
				    dahdi_qevent_lock(wc->chans[card], DAHDI_EVENT_POLARITY);
				wc->mod[card].fxo.polarity = wc->mod[card].fxo.lastpol;
		    }
		}
	}
}

static inline void pso800_proslic_check_hook(struct pso800 *wc, int card)
{
	char res;
	int hook;

	/* For some reason we have to debounce the
	   hook detector.  */

	res = wc->reg0shadow[card];
	hook = (res & 1);
	if (hook != wc->mod[card].fxs.lastrxhook) {
		/* Reset the debounce (must be multiple of 4ms) */
		wc->mod[card].fxs.debounce = 8 * (4 * 8);
#if 0
		printk("Resetting debounce card %d hook %d, %d\n", card, hook, wc->mod[card].fxs.debounce);
#endif
	} else {
		if (wc->mod[card].fxs.debounce > 0) {
			wc->mod[card].fxs.debounce-= 16 * DAHDI_CHUNKSIZE;
#if 0
			printk("Sustaining hook %d, %d\n", hook, wc->mod[card].fxs.debounce);
#endif
			if (!wc->mod[card].fxs.debounce) {
#if 0
				printk("Counted down debounce, newhook: %d...\n", hook);
#endif
				wc->mod[card].fxs.debouncehook = hook;
			}
			if (!wc->mod[card].fxs.oldrxhook && wc->mod[card].fxs.debouncehook) {
				/* Off hook */
#if 1
				if (debug)
#endif				
					printk("pso800: Card %d Going off hook\n", card);
				dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_OFFHOOK);
				if (robust)
					pso800_init_proslic(wc, card, 1, 0, 1);
				wc->mod[card].fxs.oldrxhook = 1;
			
			} else if (wc->mod[card].fxs.oldrxhook && !wc->mod[card].fxs.debouncehook) {
				/* On hook */
#if 1
				if (debug)
#endif				
					printk("pso800: Card %d Going on hook\n", card);
				dahdi_hooksig(wc->chans[card], DAHDI_RXSIG_ONHOOK);
				wc->mod[card].fxs.oldrxhook = 0;
			}
		}
	}
	wc->mod[card].fxs.lastrxhook = hook;
}

DAHDI_IRQ_HANDLER(pso800_interrupt)
{
	struct pso800 *wc = dev_id;
	int x;
	int mode;
	static unsigned int val, cfg;	
	unsigned int cnt_irq_misses;

        /* retorna se interrupcao nao foi habilitada */
	cfg = *(wc->pci_vm_addr+0x13);
	if((cfg&0x0040) == 0) {
		return IRQ_NONE;
	}
	/* retorna se interrupcao nao esta ativa */
	if((cfg&0x0004) == 0) {
		return IRQ_NONE;
	}
	/* ack da interrupcao */
	*(wc->pci_vm_addr+0x13) = (cfg|0x00000400);
	val = *(wc->ioaddr + WC_CLEAR_IRQ_REG);

	/* conta interrupcoes perdidas */
	if (wc->flag_1st_irq > 0) {
		// nao considera as primeiras passagens pela rotina
		cnt_irq_misses = (*(wc->ioaddr+WC_CNT_IRQ_REG));
		// so considera int. para o cartao
		if(cnt_irq_misses) {
			wc->flag_1st_irq--;
			*(wc->ioaddr+WC_CNT_IRQ_REG)=0;
			wc->span.irqmisses = 0;
		}
	} else {
		// neste registro da FPGA temos o numero de interrupcoes que aconteceram
		// desde o ultimo reset do contador de interrupcoes. O normal eh ler 1.
		cnt_irq_misses = (*(wc->ioaddr+WC_CNT_IRQ_REG));
		// Se for zero significa que a interrupcao nao foi gerada pelo nosso cartao
		if(cnt_irq_misses == 0) {
			if(debug) printk("Interrupcao gerada mas nao pela FPGA?!\n");
			return IRQ_NONE;

		}
		// reseta o contador
		*(wc->ioaddr+WC_CNT_IRQ_REG)=0;
		// soma interrupcoes perdidas
		wc->span.irqmisses += (cnt_irq_misses-1);
	}
		
	for (x=0;x<NUM_CARDS;x++) {
		if (wc->cardflag & (1 << x) && (wc->modtype[x] == MOD_TYPE_FXS)) {
			if (wc->mod[x].fxs.lasttxhook == 0x4) {
				/* RINGing, prepare for OHT */
				wc->mod[x].fxs.ohttimer = OHT_TIMER << 3;
				if (reversepolarity)
					wc->mod[x].fxs.idletxhookstate = 0x6;	/* OHT mode when idle */
				else
					wc->mod[x].fxs.idletxhookstate = 0x2; 
			} else {
				if (wc->mod[x].fxs.ohttimer) {
					wc->mod[x].fxs.ohttimer-= DAHDI_CHUNKSIZE;
					if (!wc->mod[x].fxs.ohttimer) {
						if (reversepolarity)
							wc->mod[x].fxs.idletxhookstate = 0x5;	/* Switch to active */
						else
							wc->mod[x].fxs.idletxhookstate = 0x1;
						if ((wc->mod[x].fxs.lasttxhook == 0x2) || (wc->mod[x].fxs.lasttxhook == 0x6)) {
							/* Apply the change if appropriate */
							if (reversepolarity) 
								wc->mod[x].fxs.lasttxhook = 0x5;
							else
								wc->mod[x].fxs.lasttxhook = 0x1;
							pso800_setreg(wc, x, 64, wc->mod[x].fxs.lasttxhook);
						}
					}
				}
			}
		}
	}

	wc->intcount++;
	x = wc->intcount & 0x3;
	mode = wc->intcount & 0xc;
	while(x<NUM_CARDS) {
		if (wc->cardflag & (1 << x)) {
			switch(mode) {
			case 0:
				/* Rest */
				break;
			case 4:
				/* Read first shadow reg */
				if (wc->modtype[x] == MOD_TYPE_FXS)
					wc->reg0shadow[x] = pso800_getreg(wc, x, 68);
				else if (wc->modtype[x] == MOD_TYPE_FXO)
					wc->reg0shadow[x] = pso800_getreg(wc, x, 5);
				break;
			case 8:
				/* Read second shadow reg */
				if (wc->modtype[x] == MOD_TYPE_FXS)
					wc->reg1shadow[x] = pso800_getreg(wc, x, 64);
				else if (wc->modtype[x] == MOD_TYPE_FXO)
					wc->reg1shadow[x] = pso800_getreg(wc, x, 29);
				break;
			case 12:
				/* Perform processing */
				if (wc->modtype[x] == MOD_TYPE_FXS) {
					pso800_proslic_check_hook(wc, x);
					if (!(wc->intcount & 0xf0))
						pso800_proslic_recheck_sanity(wc, x);
				} else if (wc->modtype[x] == MOD_TYPE_FXO) {
					pso800_voicedaa_check_hook(wc, x);
				}
				break;
			}
		}
		x += NUM_MOD_CHS;
	}
	if (!(wc->intcount % 10000)) {
		/* Accept an alarm once per 10 seconds */
		for (x=0;x<NUM_CARDS;x++) 
			if (wc->modtype[x] == MOD_TYPE_FXS) {
				if (wc->mod[x].fxs.palarms)
					wc->mod[x].fxs.palarms--;
			}
	}
	if (!(wc->intcount % 1000)) {
		for (x=0;x<NUM_CARDS;x++) {
			if(wc->cardflag & (1 << x)) {
				if(wc->mod[x].fxo.offhook)
					wc->led[x] = WC_LED_ON;
				else
					wc->led[x] = WC_LED_BLINK;
			}
		}
		for (x=0;x<NUM_CARDS;x+=2) {
			if((wc->led[x] == WC_LED_ON) || (wc->led[x+1] == WC_LED_ON))
				__pso800_set_led(wc, x/2);
			else if((wc->led[x] == WC_LED_BLINK) || (wc->led[x+1] == WC_LED_BLINK))
				__pso800_blink_led(wc, x/2);
		}
	}
	pso800_receiveprep(wc);
	pso800_transmitprep(wc);

	return IRQ_RETVAL(1);
}

static int pso800_voicedaa_insane(struct pso800 *wc, int card)
{
	int blah;

	blah = pso800_getreg(wc, card, 2);
	if (blah != 0x3)
		return -2;
	blah = pso800_getreg(wc, card, 11);
	if (debug)
		printk("VoiceDAA System: %02x\n", blah & 0xf);
	return 0;
}

static int pso800_proslic_insane(struct pso800 *wc, int card)
{
	int blah,insane_report;
	insane_report=0;

	blah = pso800_getreg(wc, card, 0);
	if (debug) 
		printk("ProSLIC on module %d, product %d, version %d\n", card, (blah & 0x30) >> 4, (blah & 0xf));

#if 0
	if ((blah & 0x30) >> 4) {
		printk("ProSLIC on module %d is not a 3210.\n", card);
		return -1;
	}
#endif
	if (((blah & 0xf) == 0) || ((blah & 0xf) == 0xf)) {
		/* SLIC not loaded */
		return -1;
	}
	if ((blah & 0xf) < 2) {
		printk("ProSLIC 3210 version %d is too old\n", blah & 0xf);
		return -1;
	}
	if (pso800_getreg(wc, card, 1) & 0x80)
		/* ProSLIC 3215, not a 3210 */
		wc->flags[card] |= FLAG_3215;

	blah = pso800_getreg(wc, card, 8);
	if (blah != 0x2) {
		printk("ProSLIC on module %d insane (1) %d should be 2\n", card, blah);
		return -1;
	} else if ( insane_report)
		printk("ProSLIC on module %d Reg 8 Reads %d Expected is 0x2\n",card,blah);

	blah = pso800_getreg(wc, card, 64);
	if (blah != 0x0) {
		printk("ProSLIC on module %d insane (2)\n", card);
		return -1;
	} else if ( insane_report)
		printk("ProSLIC on module %d Reg 64 Reads %d Expected is 0x0\n",card,blah);

	blah = pso800_getreg(wc, card, 11);
	if (blah != 0x33) {
		printk("ProSLIC on module %d insane (3)\n", card);
		return -1;
	} else if ( insane_report)
		printk("ProSLIC on module %d Reg 11 Reads %d Expected is 0x33\n",card,blah);

	/* Just be sure it's setup right. */
	pso800_setreg(wc, card, 30, 0);

	if (debug) 
		printk("ProSLIC on module %d seems sane.\n", card);
	return 0;
}

static int pso800_proslic_powerleak_test(struct pso800 *wc, int card)
{
	unsigned long origjiffies;
	unsigned char vbat;

	/* Turn off linefeed */
	pso800_setreg(wc, card, 64, 0);

	/* Power down */
	pso800_setreg(wc, card, 14, 0x10);

	/* Wait for one second */
	origjiffies = jiffies;

	while((vbat = pso800_getreg(wc, card, 82)) > 0x6) {
		if ((jiffies - origjiffies) >= (HZ/2))
			break;;
	}

	if (vbat < 0x06) {
		printk("Excessive leakage detected on module %d: %d volts (%02x) after %d ms\n", card,
		       376 * vbat / 1000, vbat, (int)((jiffies - origjiffies) * 1000 / HZ));
		return -1;
	} else if (debug) {
		printk("Post-leakage voltage: %d volts\n", 376 * vbat / 1000);
	}
	return 0;
}

static int pso800_powerup_proslic(struct pso800 *wc, int card, int fast)
{
	unsigned char vbat;
	unsigned long origjiffies;
	int lim;

	/* Set period of DC-DC converter to 1/64 khz */
	pso800_setreg(wc, card, 92, 0xff /* was 0xff */);

	/* Wait for VBat to powerup */
	origjiffies = jiffies;

	/* Disable powerdown */
	pso800_setreg(wc, card, 14, 0);

	/* If fast, don't bother checking anymore */
	if (fast)
		return 0;

	while((vbat = pso800_getreg(wc, card, 82)) < 0xc0) {
		/* Wait no more than 500ms */
		if ((jiffies - origjiffies) > HZ/2) {
			break;
		}
	}

	if (vbat < 0xc0) {
		if (wc->proslic_power == PROSLIC_POWER_UNKNOWN)
				 printk("ProSLIC on module %d failed to powerup within %d ms (%d mV only)\n\n -- DID YOU REMEMBER TO PLUG IN THE HD POWER CABLE TO THE TDM400P??\n",
					card, (int)(((jiffies - origjiffies) * 1000 / HZ)),
					vbat * 375);
		wc->proslic_power = PROSLIC_POWER_WARNED;
		return -1;
	} else if (debug) {
		printk("ProSLIC on module %d powered up to -%d volts (%02x) in %d ms\n",
		       card, vbat * 376 / 1000, vbat, (int)(((jiffies - origjiffies) * 1000 / HZ)));
	}
	wc->proslic_power = PROSLIC_POWER_ON;

        /* Proslic max allowed loop current, reg 71 LOOP_I_LIMIT */
        /* If out of range, just set it to the default value     */
        lim = (loopcurrent - 20) / 3;
        if ( loopcurrent > 41 ) {
                lim = 0;
                if (debug)
                        printk("Loop current out of range! Setting to default 20mA!\n");
        }
        else if (debug)
                        printk("Loop current set to %dmA!\n",(lim*3)+20);
        pso800_setreg(wc,card,LOOP_I_LIMIT,lim);

	/* Engage DC-DC converter */
	pso800_setreg(wc, card, 93, 0x19 /* was 0x19 */);
#if 0
	origjiffies = jiffies;
	while(0x80 & pso800_getreg(wc, card, 93)) {
		if ((jiffies - origjiffies) > 2 * HZ) {
			printk("Timeout waiting for DC-DC calibration on module %d\n", card);
			return -1;
		}
	}

#if 0
	/* Wait a full two seconds */
	while((jiffies - origjiffies) < 2 * HZ);

	/* Just check to be sure */
	vbat = pso800_getreg(wc, card, 82);
	printk("ProSLIC on module %d powered up to -%d volts (%02x) in %d ms\n",
		       card, vbat * 376 / 1000, vbat, (int)(((jiffies - origjiffies) * 1000 / HZ)));
#endif
#endif
	return 0;

}

static int pso800_proslic_manual_calibrate(struct pso800 *wc, int card){
	unsigned long origjiffies;
	unsigned char i;

	pso800_setreg(wc, card, 21, 0);//(0)  Disable all interupts in DR21
	pso800_setreg(wc, card, 22, 0);//(0)Disable all interupts in DR21
	pso800_setreg(wc, card, 23, 0);//(0)Disable all interupts in DR21
	pso800_setreg(wc, card, 64, 0);//(0)

	pso800_setreg(wc, card, 97, 0x18); //(0x18)Calibrations without the ADC and DAC offset and without common mode calibration.
	pso800_setreg(wc, card, 96, 0x47); //(0x47)	Calibrate common mode and differential DAC mode DAC + ILIM

	origjiffies=jiffies;
	while( pso800_getreg(wc,card,96)!=0 ){
		if((jiffies-origjiffies)>80)
			return -1;
	}
//Initialized DR 98 and 99 to get consistant results.
// 98 and 99 are the results registers and the search should have same intial conditions.

/*******************************The following is the manual gain mismatch calibration****************************/
/*******************************This is also available as a function *******************************************/
	// Delay 10ms
	origjiffies=jiffies; 
	while((jiffies-origjiffies)<1);
	pso800_proslic_setreg_indirect(wc, card, 88,0);
	pso800_proslic_setreg_indirect(wc,card,89,0);
	pso800_proslic_setreg_indirect(wc,card,90,0);
	pso800_proslic_setreg_indirect(wc,card,91,0);
	pso800_proslic_setreg_indirect(wc,card,92,0);
	pso800_proslic_setreg_indirect(wc,card,93,0);

	pso800_setreg(wc, card, 98,0x10); // This is necessary if the calibration occurs other than at reset time
	pso800_setreg(wc, card, 99,0x10);

	for ( i=0x1f; i>0; i--)
	{
		pso800_setreg(wc, card, 98,i);
		origjiffies=jiffies; 
		while((jiffies-origjiffies)<4);
		if((pso800_getreg(wc,card,88)) == 0)
			break;
	} // for

	for ( i=0x1f; i>0; i--)
	{
		pso800_setreg(wc, card, 99,i);
		origjiffies=jiffies; 
		while((jiffies-origjiffies)<4);
		if((pso800_getreg(wc,card,89)) == 0)
			break;
	}//for

/*******************************The preceding is the manual gain mismatch calibration****************************/
/**********************************The following is the longitudinal Balance Cal***********************************/
	pso800_setreg(wc,card,64,1);
	while((jiffies-origjiffies)<10); // Sleep 100?

	pso800_setreg(wc, card, 64, 0);
	pso800_setreg(wc, card, 23, 0x4);  // enable interrupt for the balance Cal
	pso800_setreg(wc, card, 97, 0x1); // this is a singular calibration bit for longitudinal calibration
	pso800_setreg(wc, card, 96,0x40);

	pso800_getreg(wc,card,96); /* Read Reg 96 just cause */

	pso800_setreg(wc, card, 21, 0xFF);
	pso800_setreg(wc, card, 22, 0xFF);
	pso800_setreg(wc, card, 23, 0xFF);

	/**The preceding is the longitudinal Balance Cal***/
	return(0);

}
#if 1
static int pso800_proslic_calibrate(struct pso800 *wc, int card)
{
	unsigned long origjiffies;
	int x;
	/* Perform all calibrations */
	pso800_setreg(wc, card, 97, 0x1f);
	
	/* Begin, no speedup */
	pso800_setreg(wc, card, 96, 0x5f);

	/* Wait for it to finish */
	origjiffies = jiffies;
	while(pso800_getreg(wc, card, 96)) {
		if ((jiffies - origjiffies) > 2 * HZ) {
			printk("Timeout waiting for calibration of module %d\n", card);
			return -1;
		}
	}
	
	if (debug) {
		/* Print calibration parameters */
		printk("Calibration Vector Regs 98 - 107: \n");
		for (x=98;x<108;x++) {
			printk("%d: %02x\n", x, pso800_getreg(wc, card, x));
		}
	}
	return 0;
}
#endif

static int pso800_init_voicedaa(struct pso800 *wc, int card, int fast, int manual, int sane)
{
	unsigned char reg16=0, reg17=0, reg26=0, reg30=0, reg31=0;
	long newjiffies;
	wc->modtype[card] = MOD_TYPE_FXO;
	/* Sanity check the ProSLIC */

	reset_spi(wc, card);
	if (!sane && pso800_voicedaa_insane(wc, card))
		return -2;

	/* Software reset */
	pso800_setreg(wc, card, 1, 0x80);

	/* Wait just a bit */
	wait_just_a_bit(HZ/10);

	/* Enable PCM, ulaw */
	if (alawoverride)
		pso800_setreg(wc, card, 33, 0x20);
	else
		pso800_setreg(wc, card, 33, 0x28);

	/* Set On-hook speed, Ringer impedence, and ringer threshold */
	/*reg16 |= (fxo_modes[_opermode].ohs << 6);
	reg16 |= (fxo_modes[_opermode].rz << 1);
	reg16 |= (fxo_modes[_opermode].rt);*/
	reg16 |= (channels[card].ohs << 6);
	reg16 |= (channels[card].rz << 1);
	reg16 |= (channels[card].rt);
	pso800_setreg(wc, card, 16, reg16);

	/* Cont. set ringer threshold */
	//reg17 |= (fxo_modes[_opermode].rt2 << 4);
	reg17 |= (channels[card].rt2 << 4);
	pso800_setreg(wc, card, 17, reg17);

	/* Set DC Termination:
	   Tip/Ring voltage adjust, minimum operational current, current limitation */
	/*reg26 |= (fxo_modes[_opermode].dcv << 6);
	reg26 |= (fxo_modes[_opermode].mini << 4);
	reg26 |= (fxo_modes[_opermode].ilim << 1);*/
	reg26 |= (channels[card].dcv << 6);
	reg26 |= (channels[card].mini << 4);
	reg26 |= (channels[card].ilim << 1);

	pso800_setreg(wc, card, 26, reg26);

	/* Set AC Impedence */
	//reg30 = (fxo_modes[_opermode].acim);
	reg30 = (channels[card].acim);
	pso800_setreg(wc, card, 30, reg30);

	/* Misc. DAA parameters */
	reg31 = 0xa3;
	//reg31 |= (fxo_modes[_opermode].ohs2 << 3);
	reg31 |= (channels[card].ohs2 << 3);
	pso800_setreg(wc, card, 31, reg31);

	/* Set Transmit/Receive timeslot */
	pso800_setreg(wc, card, 34, (card & 0x03) * 8);
	pso800_setreg(wc, card, 35, 0x00);
	pso800_setreg(wc, card, 36, (card & 0x03) * 8);
	pso800_setreg(wc, card, 37, 0x00);

	/* Enable ISO-Cap */
	pso800_setreg(wc, card, 6, 0x00);

	/* Wait 1000ms for ISO-cap to come up */
	newjiffies = jiffies;
	newjiffies += 2 * HZ;
	while((jiffies < newjiffies) && !(pso800_getreg(wc, card, 11) & 0xf0))
		wait_just_a_bit(HZ/10);

	if (!(pso800_getreg(wc, card, 11) & 0xf0)) {
		printk("VoiceDAA did not bring up ISO link properly!\n");
		return -1;
	}
	if (debug)
		printk("ISO-Cap is now up, line side: %02x rev %02x\n", 
		       pso800_getreg(wc, card, 11) >> 4,
		       (pso800_getreg(wc, card, 13) >> 2) & 0xf);
	/* Enable on-hook line monitor */
	pso800_setreg(wc, card, 5, 0x08);

	/* NZ -- crank the tx gain up by 7 dB */
	if (!strcmp(fxo_modes[_opermode].name, "NEWZEALAND")) {
		printk("Adjusting gain\n");
		pso800_setreg(wc, card, 38, 0x7);
	}

	return 0;
		
}

static int pso800_init_proslic(struct pso800 *wc, int card, int fast, int manual, int sane)
{

	unsigned short tmp[5];
	unsigned char r19;
	int x;
	int fxsmode=0;

	/* By default, don't send on hook */
	if (reversepolarity)
		wc->mod[card].fxs.idletxhookstate = 5;
	else
		wc->mod[card].fxs.idletxhookstate = 1;

	/* Sanity check the ProSLIC */
	if (!sane && pso800_proslic_insane(wc, card))
		return -2;
		
	if (sane) {
		/* Make sure we turn off the DC->DC converter to prevent anything from blowing up */
		pso800_setreg(wc, card, 14, 0x10);
	}

	if (pso800_proslic_init_indirect_regs(wc, card)) {
		printk(KERN_INFO "Indirect Registers failed to initialize on module %d.\n", card);
		return -1;
	}

	/* Clear scratch pad area */
	pso800_proslic_setreg_indirect(wc, card, 97,0);

	/* Clear digital loopback */
	pso800_setreg(wc, card, 8, 0);

	/* Revision C optimization */
	pso800_setreg(wc, card, 108, 0xeb);

	/* Disable automatic VBat switching for safety to prevent
	   Q7 from accidently turning on and burning out. */
	pso800_setreg(wc, card, 67, 0x17);

	/* Turn off Q7 */
	pso800_setreg(wc, card, 66, 1);

	/* Flush ProSLIC digital filters by setting to clear, while
	   saving old values */
	for (x=0;x<5;x++) {
		tmp[x] = pso800_proslic_getreg_indirect(wc, card, x + 35);
		pso800_proslic_setreg_indirect(wc, card, x + 35, 0x8000);
	}

	/* Power up the DC-DC converter */
	if (pso800_powerup_proslic(wc, card, fast)) {
		printk("Unable to do INITIAL ProSLIC powerup on module %d\n", card);
		return -1;
	}

	if (!fast) {

		/* Check for power leaks */
		if (pso800_proslic_powerleak_test(wc, card)) {
			printk("ProSLIC module %d failed leakage test.  Check for short circuit\n", card);
		}
		/* Power up again */
		if (pso800_powerup_proslic(wc, card, fast)) {
			printk("Unable to do FINAL ProSLIC powerup on module %d\n", card);
			return -1;
		}
#ifndef NO_CALIBRATION
		/* Perform calibration */
		if(manual) {
			if (pso800_proslic_manual_calibrate(wc, card)) {
				//printk("Proslic failed on Manual Calibration\n");
				if (pso800_proslic_manual_calibrate(wc, card)) {
					printk("Proslic Failed on Second Attempt to Calibrate Manually. (Try -DNO_CALIBRATION in Makefile)\n");
					return -1;
				}
				printk("Proslic Passed Manual Calibration on Second Attempt\n");
			}
		}
		else {
			if(pso800_proslic_calibrate(wc, card))  {
				//printk("ProSlic died on Auto Calibration.\n");
				if (pso800_proslic_calibrate(wc, card)) {
					printk("Proslic Failed on Second Attempt to Auto Calibrate\n");
					return -1;
				}
				printk("Proslic Passed Auto Calibration on Second Attempt\n");
			}
		}
		/* Perform DC-DC calibration */
		pso800_setreg(wc, card, 93, 0x99);
		r19 = pso800_getreg(wc, card, 107);
		if ((r19 < 0x2) || (r19 > 0xd)) {
			printk("DC-DC cal has a surprising direct 107 of 0x%02x!\n", r19);
			pso800_setreg(wc, card, 107, 0x8);
		}

		/* Save calibration vectors */
		for (x=0;x<NUM_CAL_REGS;x++)
			wc->mod[card].fxs.calregs.vals[x] = pso800_getreg(wc, card, 96 + x);
#endif

	} else {
		/* Restore calibration registers */
		for (x=0;x<NUM_CAL_REGS;x++)
			pso800_setreg(wc, card, 96 + x, wc->mod[card].fxs.calregs.vals[x]);
	}
	/* Calibration complete, restore original values */
	for (x=0;x<5;x++) {
		pso800_proslic_setreg_indirect(wc, card, x + 35, tmp[x]);
	}

	if (pso800_proslic_verify_indirect_regs(wc, card)) {
		printk(KERN_INFO "Indirect Registers failed verification.\n");
		return -1;
	}


#if 0
    /* Disable Auto Power Alarm Detect and other "features" */
    pso800_setreg(wc, card, 67, 0x0e);
    blah = pso800_getreg(wc, card, 67);
#endif

#if 0
    if (pso800_proslic_setreg_indirect(wc, card, 97, 0x0)) { // Stanley: for the bad recording fix
		 printk(KERN_INFO "ProSlic IndirectReg Died.\n");
		 return -1;
	}
#endif

    if (alawoverride)
    	pso800_setreg(wc, card, 1, 0x20);
    else
    	pso800_setreg(wc, card, 1, 0x28);
 	// U-Law 8-bit interface
    pso800_setreg(wc, card, 2, (3-card) * 8);    // Tx Start count low byte  0
    pso800_setreg(wc, card, 3, 0);    // Tx Start count high byte 0
    pso800_setreg(wc, card, 4, (3-card) * 8);    // Rx Start count low byte  0
    pso800_setreg(wc, card, 5, 0);    // Rx Start count high byte 0
    pso800_setreg(wc, card, 18, 0xff);     // clear all interrupt
    pso800_setreg(wc, card, 19, 0xff);
    pso800_setreg(wc, card, 20, 0xff);
    pso800_setreg(wc, card, 73, 0x04);
	if (fxshonormode) {
		fxsmode = acim2tiss[fxo_modes[_opermode].acim];
		pso800_setreg(wc, card, 10, 0x08 | fxsmode);
		if (fxo_modes[_opermode].ring_osc)
			pso800_proslic_setreg_indirect(wc, card, 20, fxo_modes[_opermode].ring_osc);
		if (fxo_modes[_opermode].ring_x)
			pso800_proslic_setreg_indirect(wc, card, 21, fxo_modes[_opermode].ring_x);
	}
    if (lowpower)
    	pso800_setreg(wc, card, 72, 0x10);

#if 0
    pso800_setreg(wc, card, 21, 0x00); 	// enable interrupt
    pso800_setreg(wc, card, 22, 0x02); 	// Loop detection interrupt
    pso800_setreg(wc, card, 23, 0x01); 	// DTMF detection interrupt
#endif

#if 0
    /* Enable loopback */
    pso800_setreg(wc, card, 8, 0x2);
    pso800_setreg(wc, card, 14, 0x0);
    pso800_setreg(wc, card, 64, 0x0);
    pso800_setreg(wc, card, 1, 0x08);
#endif

	if (fastringer) {
		/* Speed up Ringer */
		pso800_proslic_setreg_indirect(wc, card, 20, 0x7e6d);
		pso800_proslic_setreg_indirect(wc, card, 21, 0x01b9);
		/* Beef up Ringing voltage to 89V */
		if (boostringer) {
			pso800_setreg(wc, card, 74, 0x3f);
			if (pso800_proslic_setreg_indirect(wc, card, 21, 0x247)) 
				return -1;
			printk("Boosting fast ringer on slot %d (89V peak)\n", card + 1);
		} else if (lowpower) {
			if (pso800_proslic_setreg_indirect(wc, card, 21, 0x14b)) 
				return -1;
			printk("Reducing fast ring power on slot %d (50V peak)\n", card + 1);
		} else
			printk("Speeding up ringer on slot %d (25Hz)\n", card + 1);
	} else {
		/* Beef up Ringing voltage to 89V */
		if (boostringer) {
			pso800_setreg(wc, card, 74, 0x3f);
			if (pso800_proslic_setreg_indirect(wc, card, 21, 0x1d1)) 
				return -1;
			printk("Boosting ringer on slot %d (89V peak)\n", card + 1);
		} else if (lowpower) {
			if (pso800_proslic_setreg_indirect(wc, card, 21, 0x108)) 
				return -1;
			printk("Reducing ring power on slot %d (50V peak)\n", card + 1);
		}
	}
	pso800_setreg(wc, card, 64, 0x01);
	return 0;
}


#ifdef CONFIG_PROC_FS
int pso800_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	struct pso800 *wc = (struct pso800 *)data;
    int len = 0;
	
	if(!wc) {
	    *eof = 1;
	    return len;
	}

	len += sprintf(buf+len, wc->span.desc);
	len += sprintf(buf+len, "\n");
	len += sprintf(buf+len, "Firmware Version: %x.%x\n", (wc->fpgaver)>>8, (wc->fpgaver)&0xff);
	len += sprintf(buf+len, "Serial Number: %d\n", pso800_get_serial(wc->dev, wc));
	len += sprintf(buf+len, "Key In: 0x%04x\n", wc->write_key);
	len += sprintf(buf+len, "Key Out: 0x%08x%08x\n", *(wc->ioaddr + PSO_KEY_HIGH_REG), *(wc->ioaddr + PSO_KEY_LOW_REG));
	len += sprintf(buf+len, "Impedance: %d Ohm\n", (fxo_modes[_opermode].acim == 0)?600:900);
	len += sprintf(buf+len, "Ringer Threshold: %d Vrms\n", ((fxo_modes[_opermode].rt & fxo_modes[_opermode].rt2)?41:16) + 
			((fxo_modes[_opermode].rt | fxo_modes[_opermode].rt2)?4:0));
	len += sprintf(buf+len, "Gain Rx: %d dB\n", fxo_modes[_opermode].gain_rx);

    *eof = 1;
    return len;
}

int pso800_write_proc(struct file *file, const char *buffer, unsigned long count, void *data)
{
	#define KEY_STRING_SIZE	8
	struct pso800 *wc = (struct pso800 *)data;
	char key_string[KEY_STRING_SIZE];
	unsigned int serial;

	if(copy_from_user(key_string, buffer, KEY_STRING_SIZE)) {
		return -EFAULT;
	}
	key_string[6] = '\0';
	
	serial = pso800_get_serial(wc->dev, wc);
	sscanf(key_string, "%04x", &wc->write_key);
	//printk("pso800: using key = 0x%04x in card with SN = %d\n", wc->write_key, serial);
	
	*(wc->ioaddr + PSO_USERRAND_REG) = PSO_USERRAND_ENABLE | wc->write_key;
	*(wc->ioaddr + PSO_SERIAL_REG) = serial;
	udelay(20);
	
	return count;
}
#endif


static int pso800_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
	struct pso800_stats stats;
	struct pso800_regs regs;
	struct pso800_regop regop;
	struct pso800_echo_coefs echoregs;
	struct pso800 *wc = chan->pvt;
	struct pso800_keyregs keyregs;
	int x;
	switch (cmd) {
	case DAHDI_ONHOOKTRANSFER:
		if (wc->modtype[chan->chanpos - 1] != MOD_TYPE_FXS)
			return -EINVAL;
		if (get_user(x, (int *)data))
			return -EFAULT;
		wc->mod[chan->chanpos - 1].fxs.ohttimer = x << 3;
		if (reversepolarity)
			wc->mod[chan->chanpos - 1].fxs.idletxhookstate = 0x6;	/* OHT mode when idle */
		else
			wc->mod[chan->chanpos - 1].fxs.idletxhookstate = 0x2;
		if (wc->mod[chan->chanpos - 1].fxs.lasttxhook == 0x1) {
				/* Apply the change if appropriate */
				if (reversepolarity)
					wc->mod[chan->chanpos - 1].fxs.lasttxhook = 0x6;
				else
					wc->mod[chan->chanpos - 1].fxs.lasttxhook = 0x2;
				pso800_setreg(wc, chan->chanpos - 1, 64, wc->mod[chan->chanpos - 1].fxs.lasttxhook);
		}
		break;
	case DAHDI_SETPOLARITY:
		if (get_user(x, (int *)data))
			return -EFAULT;
		if (wc->modtype[chan->chanpos - 1] != MOD_TYPE_FXS)
			return -EINVAL;
		/* Can't change polarity while ringing or when open */
		if ((wc->mod[chan->chanpos -1 ].fxs.lasttxhook == 0x04) ||
		    (wc->mod[chan->chanpos -1 ].fxs.lasttxhook == 0x00))
			return -EINVAL;

		if ((x && !reversepolarity) || (!x && reversepolarity))
			wc->mod[chan->chanpos - 1].fxs.lasttxhook |= 0x04;
		else
			wc->mod[chan->chanpos - 1].fxs.lasttxhook &= ~0x04;
		pso800_setreg(wc, chan->chanpos - 1, 64, wc->mod[chan->chanpos - 1].fxs.lasttxhook);
		break;
	case WCTDM_GET_STATS:
		if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXS) {
			stats.tipvolt = pso800_getreg(wc, chan->chanpos - 1, 80) * -376;
			stats.ringvolt = pso800_getreg(wc, chan->chanpos - 1, 81) * -376;
			stats.batvolt = pso800_getreg(wc, chan->chanpos - 1, 82) * -376;
		} else if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXO) {
			stats.tipvolt = (signed char)pso800_getreg(wc, chan->chanpos - 1, 29) * 1000;
			stats.ringvolt = (signed char)pso800_getreg(wc, chan->chanpos - 1, 29) * 1000;
			stats.batvolt = (signed char)pso800_getreg(wc, chan->chanpos - 1, 29) * 1000;
		} else 
			return -EINVAL;
		if (copy_to_user((struct pso800_stats *)data, &stats, sizeof(stats)))
			return -EFAULT;
		break;
	case WCTDM_GET_REGS:
		if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXS) {
			for (x=0;x<NUM_INDIRECT_REGS;x++)
				regs.indirect[x] = pso800_proslic_getreg_indirect(wc, chan->chanpos -1, x);
			for (x=0;x<NUM_REGS;x++)
				regs.direct[x] = pso800_getreg(wc, chan->chanpos - 1, x);
		} else {
			memset(&regs, 0, sizeof(regs));
			for (x=0;x<NUM_FXO_REGS;x++)
				regs.direct[x] = pso800_getreg(wc, chan->chanpos - 1, x);
		}
		if (copy_to_user((struct pso800_regs *)data, &regs, sizeof(regs)))
			return -EFAULT;
		break;
	case WCTDM_SET_REG:
		if (copy_from_user(&regop, (struct pso800_regop *)data, sizeof(regop)))
			return -EFAULT;
		if (regop.indirect) {
			if (wc->modtype[chan->chanpos - 1] != MOD_TYPE_FXS)
				return -EINVAL;
			printk("Setting indirect %d to 0x%04x on %d\n", regop.reg, regop.val, chan->chanpos);
			pso800_proslic_setreg_indirect(wc, chan->chanpos - 1, regop.reg, regop.val);
		} else {
			regop.val &= 0xff;
			printk("Setting direct %d to %04x on %d\n", regop.reg, regop.val, chan->chanpos);
			pso800_setreg(wc, chan->chanpos - 1, regop.reg, regop.val);
		}
		break;
	case WCTDM_SET_ECHOTUNE:
		printk("-- Setting echo registers: \n");
		if (copy_from_user(&echoregs, (struct pso800_echo_coefs*)data, sizeof(echoregs)))
			return -EFAULT;

		if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXO) {
			/* Set the ACIM register */
			pso800_setreg(wc, chan->chanpos - 1, 30, echoregs.acim);

			/* Set the digital echo canceller registers */
			pso800_setreg(wc, chan->chanpos - 1, 45, echoregs.coef1);
			pso800_setreg(wc, chan->chanpos - 1, 46, echoregs.coef2);
			pso800_setreg(wc, chan->chanpos - 1, 47, echoregs.coef3);
			pso800_setreg(wc, chan->chanpos - 1, 48, echoregs.coef4);
			pso800_setreg(wc, chan->chanpos - 1, 49, echoregs.coef5);
			pso800_setreg(wc, chan->chanpos - 1, 50, echoregs.coef6);
			pso800_setreg(wc, chan->chanpos - 1, 51, echoregs.coef7);
			pso800_setreg(wc, chan->chanpos - 1, 52, echoregs.coef8);
                        
			printk("-- Set echo registers successfully\n");

			break;
		} else {
			return -EINVAL;

		}
		break;
	case PSO800_GET_KEY:	
		if (copy_from_user(&keyregs, (struct pso800_keyregs *)data, sizeof(struct pso800_keyregs)))
			return -EFAULT;
		keyregs.serial = pso800_get_serial(wc->dev, wc);
		keyregs.date = pso800_e2p_date(wc->dev, wc);
		*(wc->ioaddr + PSO_USERRAND_REG) = PSO_USERRAND_ENABLE | keyregs.key_low;
		*(wc->ioaddr + PSO_SERIAL_REG) = keyregs.serial;
		udelay(200);
		keyregs.key_high = *(wc->ioaddr + PSO_KEY_HIGH_REG);
		keyregs.key_low = *(wc->ioaddr + PSO_KEY_LOW_REG);
		printk("serial number = %d\n", keyregs.serial);
		printk("key = 0x%08x%08x\n", keyregs.key_high, keyregs.key_low);
		if (copy_to_user((unsigned long *)data, &keyregs, sizeof(struct pso800_keyregs)))
			return -EFAULT;
		break;
	default:
		return -ENOTTY;
	}
	return 0;

}

static int pso800_open(struct dahdi_chan *chan)
{
	struct pso800 *wc = chan->pvt;
	if (!(wc->cardflag & (1 << (chan->chanpos - 1))))
		return -ENODEV;
	if (wc->dead)
		return -ENODEV;
	wc->usecount++;

	return 0;
}

static int pso800_watchdog(struct dahdi_span *span, int event)
{

	return 0;
}

static int pso800_close(struct dahdi_chan *chan)
{
	struct pso800 *wc = chan->pvt;
	wc->usecount--;

	if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXS) {
		if (reversepolarity)
			wc->mod[chan->chanpos - 1].fxs.idletxhookstate = 5;
		else
			wc->mod[chan->chanpos - 1].fxs.idletxhookstate = 1;
	}
	/* If we're dead, release us now */
	if (!wc->usecount && wc->dead) 
		pso800_release(wc);
	return 0;
}

static int pso800_hooksig(struct dahdi_chan *chan, enum dahdi_txsig txsig)
{
	struct pso800 *wc = chan->pvt;
	int reg=0;
	if (wc->modtype[chan->chanpos - 1] == MOD_TYPE_FXO) {
		/* XXX Enable hooksig for FXO XXX */
		switch(txsig) {
		case DAHDI_TXSIG_START:
		case DAHDI_TXSIG_OFFHOOK:
			wc->mod[chan->chanpos - 1].fxo.offhook = 1;
			pso800_setreg(wc, chan->chanpos - 1, 5, 0x9);
			if (debug)
				printk("DAHDI_TXSIG_OFFHOOK - chan = %d\n", chan->chanpos);
			break;
		case DAHDI_TXSIG_ONHOOK:
#ifdef FAKE_POLARITY
			wc->mod[chan->chanpos - 1].fxo.cidtimer = wc->intcount;
#endif
			wc->mod[chan->chanpos - 1].fxo.offhook = 0;
			pso800_setreg(wc, chan->chanpos - 1, 5, 0x8);
			if (debug)
				printk("DAHDI_TXSIG_ONHOOK - chan = %d\n", chan->chanpos);
			break;
		default:
			printk("wcfxo: Can't set tx state to %d\n", txsig);
		}
	} else {
		switch(txsig) {
		case DAHDI_TXSIG_ONHOOK:
			switch(chan->sig) {
			case DAHDI_SIG_EM:
			case DAHDI_SIG_FXOKS:
			case DAHDI_SIG_FXOLS:
				wc->mod[chan->chanpos-1].fxs.lasttxhook = wc->mod[chan->chanpos-1].fxs.idletxhookstate;
				break;
			case DAHDI_SIG_FXOGS:
				wc->mod[chan->chanpos-1].fxs.lasttxhook = 3;
				break;
			}
			break;
		case DAHDI_TXSIG_OFFHOOK:
			switch(chan->sig) {
			case DAHDI_SIG_EM:
				wc->mod[chan->chanpos-1].fxs.lasttxhook = 5;
				break;
			default:
				wc->mod[chan->chanpos-1].fxs.lasttxhook = wc->mod[chan->chanpos-1].fxs.idletxhookstate;
				break;
			}
			break;
		case DAHDI_TXSIG_START:
			wc->mod[chan->chanpos-1].fxs.lasttxhook = 4;
			break;
		case DAHDI_TXSIG_KEWL:
			wc->mod[chan->chanpos-1].fxs.lasttxhook = 0;
			break;
		default:
			printk("pso800: Can't set tx state to %d\n", txsig);
		}
		if (debug)
			printk("Setting FXS hook state to %d (%02x)\n", txsig, reg);

#if 1
		pso800_setreg(wc, chan->chanpos - 1, 64, wc->mod[chan->chanpos-1].fxs.lasttxhook);
#endif
	}
	return 0;
}

static int pso800_initialize(struct pso800 *wc)
{
	int x;
	char type_str[16];

	/* dahdiata stuff */
	sprintf(wc->span.name, "PSO800/%d", wc->pos);
	sprintf(wc->span.desc, "%s Card %d", wc->variety, wc->pos + 1);
	if (alawoverride) {
		printk("ALAW override parameter detected.  Device will be operating in ALAW\n");
		wc->span.deflaw = DAHDI_LAW_ALAW;
	} else
		wc->span.deflaw = DAHDI_LAW_MULAW;
	for (x = 0; x < NUM_CARDS; x++) {
		if(!((wc->cardflag)&(1<<x)))
			sprintf(type_str, "N/C");
		else if(wc->modtype[x] == MOD_TYPE_FXO)
			sprintf(type_str, "FXO");
		else if(wc->modtype[x] == MOD_TYPE_FXS)
			sprintf(type_str, "FXS");
		else
			sprintf(type_str, "???");
		sprintf(wc->chans[x]->name, "PSO800/%d/%d %s", wc->pos, x, type_str);
				
		wc->chans[x]->sigcap = DAHDI_SIG_FXOKS | DAHDI_SIG_FXOLS | DAHDI_SIG_FXOGS | DAHDI_SIG_SF | DAHDI_SIG_EM | DAHDI_SIG_CLEAR;
		wc->chans[x]->sigcap |= DAHDI_SIG_FXSKS | DAHDI_SIG_FXSLS | DAHDI_SIG_SF | DAHDI_SIG_CLEAR;
		wc->chans[x]->chanpos = x+1;
		wc->chans[x]->pvt = wc;
	}
	wc->span.chans = wc->chans;
	wc->span.channels = NUM_CARDS;
	wc->span.hooksig = pso800_hooksig;
	wc->span.open = pso800_open;
	wc->span.close = pso800_close;
	wc->span.flags = DAHDI_FLAG_RBS;
	wc->span.ioctl = pso800_ioctl;
	wc->span.watchdog = pso800_watchdog;
	init_waitqueue_head(&wc->span.maintq);

	wc->span.pvt = wc;
	if (dahdi_register(&wc->span, 0)) {
		printk("Unable to register span with dahdi\n");
		return -1;
	}
	return 0;
}

static void pso800_post_initialize(struct pso800 *wc)
{
	int x;
	/* Finalize signalling  */
	for (x = 0; x < NUM_CARDS; x++) {
		if (wc->cardflag & (1 << x)) {
			if (wc->modtype[x] == MOD_TYPE_FXO)
				wc->chans[x]->sigcap = DAHDI_SIG_FXSKS | DAHDI_SIG_FXSLS | DAHDI_SIG_SF | DAHDI_SIG_CLEAR;
			else if (wc->modtype[x] == MOD_TYPE_FXS)
				wc->chans[x]->sigcap = DAHDI_SIG_FXOKS | DAHDI_SIG_FXOLS | DAHDI_SIG_FXOGS | DAHDI_SIG_SF | DAHDI_SIG_EM | DAHDI_SIG_CLEAR;
		}
	}
}

static int pso800_hardware_init(struct pso800 *wc)
{
	unsigned char x;

	/* Setup our proper outputs for when we switch for our "serial" port */
	wc->ios = BIT_CS | BIT_SCLK | BIT_SDI;
	*(wc->ioaddr + WC_SPI_REG) = wc->ios;

	/* mantem pico em reset */
	*(wc->ioaddr + WC_PICO_REG) = 0x00;

	// Hardware reset	
	*(wc->ioaddr + WC_RST_REG) = 0x00;
	/* Wait 1/4 of a sec */
	wait_just_a_bit(HZ/4);
	*(wc->ioaddr + WC_RST_REG) = 0x01;	
	/* Wait 1/4 of a second more */
	wait_just_a_bit(HZ/4);

#if 0
	for (x = 0; x < NUM_CARDS; x++) {
		int sane=0,ret=0,readi=0;
		// Init with Auto Calibration 
		if (!(ret=pso800_init_proslic(wc, x, 0, 0, sane))) {
			wc->cardflag |= (1 << x);
                        if (debug) {
                                readi = pso800_getreg(wc,x,LOOP_I_LIMIT);
                                printk("Proslic module %d loop current is %dmA\n",x,
                                ((readi*3)+20));
                        }
			printk("Module %d: Installed -- AUTO FXS/DPO\n",x);
		} else {
			if(ret!=-2) {
				sane=1;
				// Init with Manual Calibration
				if (!pso800_init_proslic(wc, x, 0, 1, sane)) {
					wc->cardflag |= (1 << x);
                                if (debug) {
                                        readi = pso800_getreg(wc,x,LOOP_I_LIMIT);
                                        printk("Proslic module %d loop current is %dmA\n",x,
                                        ((readi*3)+20));
                                }
					printk("Module %d: Installed -- MANUAL FXS\n",x);
				} else {
					printk("Module %d: FAILED FXS (%s)\n", x, fxshonormode ? fxo_modes[_opermode].name : "FCC");
				} 
			} else if (!(ret = pso800_init_voicedaa(wc, x, 0, 0, sane))) {
				wc->cardflag |= (1 << x);
				printk("Module %d: Installed -- AUTO FXO (%s mode)\n",x, fxo_modes[_opermode].name);
			} else {
				wc->led[x] = WC_LED_OFF;
				printk("Module %d: Not installed\n", x);
			}
		}
	}
#else
	for (x = 0; x < NUM_CARDS; x++) {
		int sane=0,ret=0;

		if (!(ret = pso800_init_voicedaa(wc, x, 0, 0, sane))) {
			wc->cardflag |= (1 << x);
			//printk("Module %d: Installed -- AUTO FXO (%s mode)\n",x, fxo_modes[_opermode].name);
			printk("Module %d: Installed -- AUTO FXO (%s mode)\n", x, channels[x].name);
		} else {
			wc->led[x] = WC_LED_OFF;
			printk("Module %d: Not installed\n", x);
		}
	}
#endif

	/* Return error if nothing initialized okay. */
	if (!wc->cardflag && !timingonly)
		return -1;

	return 0;
}


static void pso800_enable_interrupts(struct pso800 *wc)
{
	unsigned int val;
	wc->flag_1st_irq = 10;
	// Enable LINTi1 in edge triggerable mode
	*(wc->pci_vm_addr+0x13) |= 0x00000103;
	// Enable PCI interrupt pin
	*(wc->pci_vm_addr+0x13) |= 0x00000040;
	// Limpa interrupcao da FPGA para forcar borda de subida na proxima
	val = *(wc->ioaddr + WC_CLEAR_IRQ_REG);
}


static void pso800_disable_interrupts(struct pso800 *wc)	
{
	*(wc->pci_vm_addr+0x13) &= 0xFFFF0000;
}

u32 pso800_e2prom_read(struct pci_dev *dev, u16 addr)
{
	unsigned int jiffies_hold = jiffies;
	u16 data16;
	u32 data32;
	
	data16 = addr;
	pci_write_config_word(dev, PSO_E2P_PVDADDR_REG, data16);	
	while(!(data16 & PSO_E2P_OPER_FLAG)) {
		pci_read_config_word(dev, PSO_E2P_PVDADDR_REG, &data16);
		if(jiffies>(jiffies_hold+100)) {
			printk("pso800_e2prom_read: timeout!\n");
			return -1;
		}
	}
	pci_read_config_dword(dev, PSO_E2P_PVDATA_REG, &data32);
	return data32;
}

unsigned int pso800_get_serial(struct pci_dev *dev, struct pso800 *wc)
{
	unsigned int sn = pso800_e2prom_read(dev, PSO_E2P_ADDR_SERIAL);

	if(sn == 0xFFFFFFFF) return 0;
	else return sn;
}

unsigned int pso800_e2p_date(struct pci_dev *dev, struct pso800 *wc)
{
	unsigned int date = pso800_e2prom_read(dev, PSO_E2P_ADDR_DATE);
	
	if(date == 0xFFFFFFFF) return 0;
	else return date;
}

int pso800_bus_test(struct pso800 *wc)
{
	int i=0, tst_result = 0;
	unsigned int val;
	
		*(wc->ioaddr+WC_BUS_TEST_REG) = 0xAAAAAAAA;
		*wc->ioaddr = 0; // flush
		val = *(wc->ioaddr+WC_BUS_TEST_REG);
		if(val != 0xAAAAAAAA) {
			printk("Escrito 0xAAAAAAAA, lido 0x%08X! (%d)\n", val, i);
			tst_result++;
		}
		*(wc->ioaddr+WC_BUS_TEST_REG) = 0x55555555;
		*wc->ioaddr = 0; // flush
		val = *(wc->ioaddr+WC_BUS_TEST_REG);
		if(val != 0x55555555) {
			printk("Escrito 0x55555555, lido 0x%08X! (%d)\n", val, i);
			tst_result++;
		}
		*(wc->ioaddr+WC_BUS_TEST_REG) = 0xFFFFFFFF;
		*wc->ioaddr = 0; // flush
		val = *(wc->ioaddr+WC_BUS_TEST_REG);
		if(val != 0xFFFFFFFF) {
			printk("Escrito 0xFFFFFFFF, lido 0x%08X! (%d)\n", val, i);
			tst_result++;
		}	
		*(wc->ioaddr+WC_BUS_TEST_REG) = 0x00000000;
		*wc->ioaddr = 0xFFFFFFFF; // flush
		val = *(wc->ioaddr+WC_BUS_TEST_REG);
		if(val != 0x00000000) {
			printk("Escrito 0x00000000, lido 0x%08X! (%d)\n", val, i);
			tst_result++;
		}
		return tst_result;
}

static int pso800_pci_setup(struct pci_dev *dev, const struct pci_device_id *id, struct pso800 *wc)
{
	unsigned long int pci_start_hwaddr, pci_end_hwaddr, pci_len_hwaddr;
	u8 irq_num;
	u16 data;
	u32 word32;


	pd3_debugk("\n");
	/* Verifica VENDOR e DEVICE */
	pci_read_config_word(dev, PCI_VENDOR_ID, &data);
	if(data != PCI_VENDOR_ID_PLX) {
		printk("PCI_VENDOR_ID (0x%04x) diferente do esperado para o cartao (0x%04x)\n", data, PCI_VENDOR_ID_PLX);
		return -ENODEV;
	}
	pci_read_config_word(dev, PCI_DEVICE_ID, &data);
	if(data != PCI_DEVICE_ID_PLX_9030) {
		printk("PCI_DEVICE_ID (0x%04x) diferente do esperado para o cartao (0x%04x)\n", data, PCI_DEVICE_ID_PLX_9030);
		return -ENODEV;
	}

	/* imprime status e registro de comando */
	pci_read_config_word(dev, PCI_COMMAND, &data);
//	printk("registro de comando = 0x%04x\n", data);
	pci_read_config_word(dev, PCI_STATUS, &data);
//	printk("registro de status  = 0x%04x\n", data);

	/* le numero da IRQ associada */
	pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &irq_num);
	printk("irq number = 0x%02x\n", irq_num);

	/* verifica e mapeia primeiro espaco de enderecamento
	   este espaco eh acessado via espaco de memoria e da acesso
	   aos registros de configuracao local do PLX */
	pci_start_hwaddr = pci_resource_start(dev, 0);
	pci_end_hwaddr = pci_resource_end(dev, 0);
	pci_len_hwaddr = pci_end_hwaddr - pci_start_hwaddr;
	printk("pci_resource_start = 0x%08lx\n", pci_start_hwaddr);
	printk("pci_resource_end = 0x%08lx\n", pci_end_hwaddr);

	/* aloca regioes de memoria */
	if (pci_request_regions(dev, "pso800"))
			printk("pso800: Unable to request regions\n");

	/* verifica se regiao de memoria esta livre e aloca a mesma*/
/*	if(request_mem_region(pci_start_hwaddr, pci_len_hwaddr, "e1_config_space") == NULL) {
		printk("request_mem_region falhou!!!\n");
		return -ENODEV;
	}
*/

	/* faz o ioremap da regiao alocada */
	wc->pci_vm_addr = ioremap_nocache(pci_start_hwaddr, pci_len_hwaddr);
	if(wc->pci_vm_addr == NULL) {
		printk("ioremap falhou!!!\n");
		release_mem_region(pci_start_hwaddr, pci_len_hwaddr);
		return -ENODEV;
	}

	/* Agora que temos ponteiro para os registros locais de configuracao,
		ativa address space 0 e CS0 */
	*(wc->pci_vm_addr+0x0a) = 0x4090A840;		// LAS0BRD (offset = 0x28) -> insere 1 wait state em todos os ciclos
	*(wc->pci_vm_addr+0x0f) = 0x00080001;		// CS0BASE (offset = 0x3c) -> habilita CS0
	*(wc->pci_vm_addr+0x13) = 0x003F0000;
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &word32);
//	printk("PCI_BASE_ADDRESS_2 = 0x%08x\n", word32);
	printk("local cfg reg 0x00 = 0x%08x\n", *(wc->pci_vm_addr));
	
	return 0;
}

static int __devinit pso800_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int res;
	struct pso800 *wc;
	struct pso800_desc *d = (struct pso800_desc *)ent->driver_data;
	int x;
	int y;
	int btest;
	static int initd_ifaces=0;
	int cardcount = 0;	
#ifdef CONFIG_PROC_FS	
	static struct proc_dir_entry *rw_proc_file;
#endif
	
	if(!initd_ifaces){
		memset((void *)ifaces,0,(sizeof(struct pso800 *))*WC_MAX_IFACES);
		initd_ifaces=1;
	}
	for (x=0;x<WC_MAX_IFACES;x++) {
		if (!ifaces[x])
			break;
	}
	if (x >= WC_MAX_IFACES) {
		printk("Too many interfaces\n");
		return -EIO;
	}
	
	if (pci_enable_device(pdev)) {
		res = -EIO;
	} else {

		wc = kmalloc(sizeof(struct pso800), GFP_KERNEL);
		if (wc) {
			
			memset(wc, 0, sizeof(struct pso800));
			for(x=0; x < sizeof(wc->chans)/sizeof(wc->chans[0]); ++x) {
				wc->chans[x] = &wc->_chans[x];
			}

			if(pso800_pci_setup(pdev, ent, wc)<0) {
				kfree(wc);
				return -EIO;
			}
			spin_lock_init(&wc->lock);

			wc->memaddr = pci_resource_start(pdev, 2);
			wc->memlen = pci_resource_len(pdev, 2);			
			printk("pci_resource_start(dev, 2) = 0x%08lx\n", wc->memaddr);
			printk("pci_resource_len(dev, 2) = 0x%08lx\n", wc->memlen);
			// verifica se regiao de memoria esta livre e aloca a mesma
//			if(request_mem_region(wc->memaddr, wc->memlen, wc->variety) == NULL) {
//				printk("request_mem_region falhou!!!\n");
//				return -EIO;
//			}
			// faz o ioremap da regiao alocada
			wc->ioaddr = ioremap_nocache(wc->memaddr, wc->memlen);
			if(wc->ioaddr == NULL) {
				printk("ioremap falhou!!!\n");
				release_mem_region(wc->memaddr, wc->memlen);
				kfree(wc);
				return -EIO;
			} 
			
			// verifica modelo do cartao
			if(*wc->ioaddr == PSO800_CARD_ID) {
				printk("Found a PSO800 card (Intelbras card)!\n");
				// executa teste preliminar de escrita/leitura no barramento
				if(pso800_bus_test(wc)) {
					printk("ERRO NO TESTE DE BARRAMENTO!!!\n");
					kfree(wc);
					return -EIO;
				}

				wc->fpgaver = (*(wc->ioaddr+WC_CARD_TYPE_REG))&0xffff;
				/*if (wc->fpgaver > 0x0100) {
					*(wc->pci_vm_addr+0x0a) = 0x00800000;	// LAS0BRD (offset = 0x28) -> sem wait states
					printk("Using no wait states\n");
				} else {
					printk("Using 1 wait state\n");
				}*/

				// executa novo teste de barramento
				for(btest=0; btest<1000; btest++) {
					if(pso800_bus_test(wc)) {
						printk("ERRO NO TESTE DE BARRAMENTO!!!\n");
						kfree(wc);
						return -EIO;
					}
				}
				printk("Barramento OK, fpgaver = %x.%x\n", wc->fpgaver>>8, wc->fpgaver&0xff);
			} else {
				printk("Card is NOT a Intelbras PSO800 card! (ID = 0x%08x)\n", *wc->ioaddr);
				kfree(wc);
				return -EIO;
			}

#ifdef FAKE_POLARITY						
			if(dtmfthr < 2000)
				dtmfthr = 2000;
			printk("Using dtmfthr of %d\n", dtmfthr);
#endif

			wc->curcard = -1;
			wc->dev = pdev;
			wc->pos = x;
			wc->variety = d->name;
			for (y=0;y<NUM_CARDS;y++)
				wc->flags[y] = d->flags;
			/* Keep track of whether we need to free the region */
			if (request_region((unsigned long)wc->ioaddr, 0x3ff, "pso800")) 
				wc->freeregion = 1;

			/* Allocate enough memory for two dahdi chunks, receive and transmit.  Each sample uses
			   32 bits.  Allocate an extra set just for control too */
			wc->writechunk = pci_alloc_consistent(pdev, DAHDI_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, &wc->writedma);
			if (!wc->writechunk) {
				printk("pso800: Unable to allocate DMA-able memory\n");
				if (wc->freeregion)
					release_region((unsigned long)wc->ioaddr, 0x3ff);
				return -ENOMEM;
			}
			wc->readchunk = wc->writechunk + DAHDI_MAX_CHUNKSIZE * 2;	/* in doublewords */
			wc->readdma = wc->writedma + DAHDI_MAX_CHUNKSIZE * 8;		/* in bytes */

			if (pso800_hardware_init(wc)) {
				// Free Resources
				free_irq(pdev->irq, wc);
				if (wc->freeregion)
					release_region((unsigned long)wc->ioaddr, 0x3ff);
				pci_free_consistent(pdev, DAHDI_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, (void *)wc->writechunk, wc->writedma);
				pci_set_drvdata(pdev, NULL);
				dahdi_unregister(&wc->span);
				kfree(wc);
				return -EIO;

			}

			if (pso800_initialize(wc)) {
				printk("pso800: Unable to intialize FXS\n");
				/* Free Resources */
				free_irq(pdev->irq, wc);
				if (wc->freeregion)
					release_region((unsigned long)wc->ioaddr, 0x3ff);
				pci_free_consistent(pdev, DAHDI_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, (void *)wc->writechunk, wc->writedma);
				kfree(wc);
				return -EIO;
			}

			/* Keep track of which device we are */
			pci_set_drvdata(pdev, wc);
			if (request_irq(pdev->irq, pso800_interrupt, SA_SHIRQ, "pso800", wc)) {
				printk("pso800: Unable to request IRQ %d\n", pdev->irq);
				if (wc->freeregion)
					release_region((unsigned long)wc->ioaddr, 0x3ff);
				pci_free_consistent(pdev, DAHDI_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, (void *)wc->writechunk, wc->writedma);
				pci_set_drvdata(pdev, NULL);
				kfree(wc);
				return -EIO;
			}

			ifaces[wc->pos] = wc;
			
			pso800_post_initialize(wc);

			/* Enable interrupts */
			pso800_enable_interrupts(wc);
			
			/* Initialize Write/Buffers to all blank data */
			memset((void *)wc->writechunk,0,DAHDI_MAX_CHUNKSIZE * 2 * 2 * 4);

			for (x = 0; x < NUM_CARDS; x++) {
				if (wc->cardflag & (1 << x))
					cardcount++;
			}

#ifdef CONFIG_PROC_FS
			sprintf(wc->proc_filename, "intelbras/PSO800_%d", wc->pos+1);
			rw_proc_file = create_proc_entry(wc->proc_filename, 0666, NULL);
			if(rw_proc_file == NULL) {
				proc_mkdir("intelbras", NULL);
				rw_proc_file = create_proc_entry(wc->proc_filename, 0666, NULL);
				if(rw_proc_file == NULL)
					printk("Unable to create /proc/intelbras entry!!\n");
			} 
			if(rw_proc_file) {
				rw_proc_file->read_proc = pso800_read_proc;
				rw_proc_file->write_proc = pso800_write_proc;
				rw_proc_file->owner = THIS_MODULE;
				rw_proc_file->mode = S_IFREG | S_IRUGO;
				rw_proc_file->uid = 0;
				rw_proc_file->gid = 0;
				rw_proc_file->size = 0;			
				rw_proc_file->data = (void *) wc;
			}
#endif

			printk("Found a Wildcard TDM: %s (%d modules, SN = %d)\n", wc->variety, cardcount, pso800_get_serial(pdev, wc));
			res = 0;
		} else
			res = -ENOMEM;
	}
	return res;
}

static void pso800_release(struct pso800 *wc)
{
	dahdi_unregister(&wc->span);
	if (wc->freeregion)
		release_region((unsigned long)wc->ioaddr, 0x3ff);
	kfree(wc);
	printk("Freed a Wildcard\n");
}

static void __devexit pso800_remove_one(struct pci_dev *pdev)
{
	struct pso800 *wc = pci_get_drvdata(pdev);
	int x;
	if (wc) {

		/* In case hardware is still there */
		pso800_disable_interrupts(wc);
		
		for (x = 0; x < (NUM_CARDS/2); x++) {
			__pso800_clear_led(wc, x);		
		}
		
		/* Immediately free resources */
		pci_free_consistent(pdev, DAHDI_MAX_CHUNKSIZE * 2 * 2 * 2 * 4, (void *)wc->writechunk, wc->writedma);
		free_irq(pdev->irq, wc);

#ifdef CONFIG_PROC_FS
		remove_proc_entry(wc->proc_filename, NULL);
#endif

		/* Release span, possibly delayed */
		if (!wc->usecount)
			pso800_release(wc);
		else
			wc->dead = 1;
	}
}

static struct pci_device_id pso800_pci_tbl[] __devinitdata =
{
	{ PCI_VENDOR_ID_PLX, PCI_DEVICE_ID_PLX_9030, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (unsigned long) &pso800r1 },
	{ 0, }
};


MODULE_DEVICE_TABLE(pci, pso800_pci_tbl);

static struct pci_driver pso800_driver = {
	name: 	"pso800",
	probe: 	pso800_init_one,
	remove:	__devexit_p(pso800_remove_one),
	suspend: NULL,
	resume:	NULL,
	id_table: pso800_pci_tbl,
};

/* XXX UGLY XXX */
static int pso800_get_param_chan(char *str, int num) {
	int i = 0, x, gain_rx;
	char *param;

	if(str == NULL)
		str = opermode;

	while ( (param = strsep(&str, ",")) != NULL) {
		switch (i++) {
			case 0:
				for (x=0;x<(sizeof(fxo_modes) / sizeof(fxo_modes[0])); x++) {
					if (!strcmp(fxo_modes[x].name, param))
						break;
				}
				if (x < sizeof(fxo_modes) / sizeof(fxo_modes[0])) {
					_opermode = x;
				} else {
					printk("Invalid/unknown operating mode '%s' specified.  Please choose one of:\n", param);
					for (x=0;x<sizeof(fxo_modes) / sizeof(fxo_modes[0]); x++)
						printk("  %s\n", fxo_modes[x].name);
					printk("Note this option is CASE SENSITIVE!\n");
					return -ENODEV;
				}
				if (!strcmp(fxo_modes[_opermode].name, "AUSTRALIA")) {
					boostringer=1;
					fxshonormode=1;
				}
				channels[num] = fxo_modes[_opermode];
				break;
			case 1:
				if (!strcmp(param, "600"))
					channels[num].acim = 0;
				else if (!strcmp(param, "900"))
					channels[num].acim = 1;
				else {
					printk("Invalid Impedance in channel %d\n", num+1);
					return -ENODEV;
				}
				printk("Set Impedance: %s Ohm in channel %d\n", param, num+1);
				break;
			case 2:
				if (!strcmp(param, "16")) {
					channels[num].rt = 0;
				    channels[num].rt2 = 0;
				}
				else if (!strcmp(param, "20")) {
					channels[num].rt = 1;
				    channels[num].rt2 = 0;
				}
				else if (!strcmp(param, "45")) {
					channels[num].rt = 1;
				    channels[num].rt2 = 1;
				}
				else {
					printk("Invalid Ringer Threshold in channel %d\n", num+1);
					return -ENODEV;
				}
				printk("Set Ringer Threshold: %sVrms in channel %d\n", param, num+1);
				break;
			case 3:
				gain_rx = (int)simple_strtol(param, NULL, 10);
				if(gain_rx >= 0 && gain_rx <= 12) {
					channels[num].gain_rx = gain_rx;
				} else {
					printk("Invalid Gain Rx in channel %d\n", num+1);
					return -ENODEV;
				}
				printk("Set gain Rx: %ddB in channel %d\n", gain_rx, num+1);
				break;
			default:
				printk("Error in opermode param\n");
				return -ENODEV;
		}
	}
	return 1;
}

static int __init pso800_init(void)
{
	int res, i;
	
	char *chan[NUM_CARDS] = {chan1, chan2, chan3, chan4, chan5, chan6, chan7, chan8};

	printk("Intelbras PSO800 Dahdi PCI Card Driver\n");

	/* XXX UGLY XXX */
	for(i=0; i<NUM_CARDS; i++) {
		res = pso800_get_param_chan(chan[i], i);
		if(res < 0)
			return res;
	}

	res = dahdi_pci_module(&pso800_driver);
	if (res)
		return -ENODEV;
	return 0;
}

static void __exit pso800_cleanup(void)
{
	pci_unregister_driver(&pso800_driver);
}

module_param(debug, int, 0600);
module_param(loopcurrent, int, 0600);
module_param(reversepolarity, int, 0600);
module_param(robust, int, 0600);
module_param(_opermode, int, 0600);
module_param(opermode, charp, 0600);
module_param(timingonly, int, 0600);
module_param(lowpower, int, 0600);
module_param(boostringer, int, 0600);
module_param(fastringer, int, 0600);
module_param(fxshonormode, int, 0600);
module_param(battdebounce, int, 0600);
module_param(battthresh, int, 0600);
module_param(alawoverride, int, 0600);
/* XXX UGLY XXX */
module_param(chan1, charp, 0600);
module_param(chan2, charp, 0600);
module_param(chan3, charp, 0600);
module_param(chan4, charp, 0600);
module_param(chan5, charp, 0600);
module_param(chan6, charp, 0600);
module_param(chan7, charp, 0600);
module_param(chan8, charp, 0600);
#ifdef FAKE_POLARITY
module_param(dtmfthr, int, 0600);
module_param(minvoltage, int, 0600);
#endif
MODULE_DESCRIPTION("Intelbras PSO800 Dahdi PCI Card Driver");
MODULE_AUTHOR("Intelbras <rvaliati@pd3.com.br>");
#if defined(MODULE_ALIAS)
MODULE_ALIAS("pso800");
#endif
MODULE_LICENSE("GPL v2");

module_init(pso800_init);
module_exit(pso800_cleanup); 
